/////////////////////////////////////////////////////////////////////////////
//
// File: r1emu.h
//
//
// The R1-emulator - global declarations
// -------------------------------------
//
// This file defines standard types and constants for the R1-emulator.
// 
//   * Opcode2reg 
//   * Opcode1reg 
//   * Opcode0reg
//   * Opcode (union of first three)
//   * Flags
//
// (c) 1999-2000 Rune Barnkob
//
//
#ifndef R1EMU_H
#define R1EMU_H

// These typedefs generates the default sizes in the system.
// Rules are: BYTE is 16-bit unsigned (intel endian on 8-bit machines)
//            SHORTINT is 16-bit signed
//            WORD is 32-bit unsigned (pair of bytes, MSB first)
//            INT is 32-bit signed (signed WORD)
typedef unsigned short int BYTE;
typedef unsigned int WORD;
typedef signed short int SHORTINT;
typedef signed int INT;


//
// NOTICE ABOUT BYTEORDER!!!
//
// The structures are designed to be generated from least-significant-bit and
// on. If this is not the case for your compiler, you must modify the
// structures so they match. This means that, for instance, in Opcode2reg
// "dest" is the 5 least significant bit!


//*********** 2-REG OPERATIONS **********
// Opcodes for group-1 : 2-reg opcodes 
#define OC_NOP     0x00 // No-OPeration 
#define OC_ADD     0x02 // ADD, ADC 
#define OC_SUB     0x03 // SUB, SBB, CMP 
#define OC_LOGICAL 0x04 // AND, TEST, OR, XOR 
#define OC_BITTEST 0x05 // BTT, BTR - BiTTest, BitTRansfer
#define OC_BITMAN  0x06 // BTS, BTC - BiTSet, BiTClear
#define OC_MULDIV  0x07 // MUL, IMUL, DIV, IDIV - Multiplication and division (N/A) 
#define OC_SHR     0x08 // SHR, SHR imm, SAR, SAR imm 
#define OC_SHL     0x09 // SHL, SHL imm 
#define OC_ROR     0x0A // ROR, ROR imm, RCR, RCR imm 
#define OC_ROL     0x0B // ROL, ROL imm, RCL, RCL imm 
#define OC_MOVE    0x0C // Move between mem and regs 
#define OC_MOVE_H  0x0D // Move between mem and regs 
#define OC_MOVEBO  0x0E // Move base-immediate-offset, param1 chooses word-move, param2 direction (1=store) 
// Structure for 2-reg operations 
struct Opcode2reg {
	unsigned dest   : 5;
	unsigned source : 5;
	unsigned param2 : 1;
	unsigned param1 : 1;
	unsigned inst   : 4;
};

//*********** 1-REG OPERATIONS **********
// Opcodes for group-2 : 1-reg opcodes 
#define OC_CLR       0x000 // CLR - reg=0
#define OC_SET       0x001 // SET - reg=FFFFFFFFh
#define OC_NOT       0x002 // NOT - reg=!reg
#define OC_NEG       0x003 // NEG - reg=-reg
#define OC_INC       0x004 // INC - add 1 to reg
#define OC_DEC       0x005 // DEC - decrese reg with 1
#define OC_PUSH      0x006 // PUSH - push register on stack
#define OC_POP       0x007 // POP - pop register off stack
#define OC_LOADIMM   0x008 // MOVE.I - move immediate into register
#define OC_LOADIMM_H 0x009 // MOVE.IB - move immediate byte into register
// conditional and unconditional jumps - same values for 1-reg and 0-reg   
#define OC_JMP       0x00E // unconditional jump                           
#define OC_CALL      0x00F // unconditional call (IP pushed on stack)      
#define OC_JE        0x010 // jump : equal (ZF=1)                          
#define OC_JNE       0x011 //  if  : not equal (ZF=0)                      
#define OC_JC        0x012 //      : carry set (CF=1)                      
#define OC_JNC       0x013 //      : carry clear (CF=0)                    
#define OC_JS        0x014 //      : sign set / negative (SF=1)            
#define OC_JNS       0x015 //      : sign clear (SF=0)                     
#define OC_JA        0x016 //      : above (ZF or CF) = 0                  
#define OC_JAE       0x017 //      : above or equal (CF=0)                 
#define OC_JB        0x018 //      : below (CF=1)                          
#define OC_JBE       0x019 //      : below or equal (CF or ZF) = 1         
#define OC_JG        0x01A //      : greater ((SF xor OF) or ZF) = 0       
#define OC_JGE       0x01B //      : greater or equal (SF xor OF) = 0      
#define OC_JL        0x01C //      : less (SF xor OF) = 1                  
#define OC_JLE       0x01D //      : less or equal ((SF xor OF) or ZF) = 1 
#define OC_JO        0x01E //      : overflow  (OF=1)                      

// Structure for 1-reg operation 
struct Opcode1reg {
	unsigned reg      : 5;
	unsigned inst     : 5;
	unsigned reserved : 6; // =111111b 
};


//************ 0-REG OPERATIONS **********
// Opcodes for group-3 : 0-reg opcodes 
#define OC_HALT 0x01f // HALT - halt CPU until interrupt
#define OC_RET  0x000 // RET - return from subroutine (pop IP off stack)
#define OC_IRET 0x001 // IRET -  return from interrupt (pop IP and FLAGS off stack)
// Structure for 0-reg operations 
struct Opcode0reg {
	unsigned inst     : 5;
	unsigned reserved : 11; /* = 111111 11111 b */
};


//** GENERAL STRUCTURE FOR INSTRUCTION **
union Opcode {
	BYTE opcode;
	struct Opcode2reg twoReg;
	struct Opcode1reg oneReg;
	struct Opcode0reg zeroReg;
};


//** CPU FLAGS **
struct Flags {
	unsigned zero       : 1; // operand was zero 
	unsigned carry      : 1; // Carry out-of-addition 
	unsigned sign       : 1; // Sign of result 
	unsigned overflow   : 1; // Set if arithmetic op gives overflow 
	unsigned bytearith  : 1; // Set if arithmetic works on half-bytes only 
	unsigned reserved3  : 3; // reserved = 0
	unsigned intrenable : 1; // Interrupt-mask, 0=no int, 1=int enabled
  unsigned singlestep : 1; // Interrupt on single step
  unsigned intoverflow: 1; // Interrupt on overflow
	unsigned reserved2  : 5; // reserved = 0
  unsigned reserved1  : 8; // reserved = 0
	unsigned reserved0  : 5; // reserved = 0
	unsigned paging     : 1; // Paging enabled 
	unsigned supervisor : 2; // Supervisor level 
};

//** DEFINED INTERRUPTS **
#define INT_NMI     0
#define INT_DBZ     1
#define INT_OPCODE  2
#define INT_RAMERR  3
#define INT_BRKPNT  4
#define INT_PFAULT  5
#define INT_OVRFLOW 8
#define INT_IEXT0   9
#define INT_TIMER0 10
#define INT_IEXT1  11
#define INT_TIMER1 12
#define INT_IEXT2  13
#define INT_TIMER2 14
#define INT_TIMER3 15

#endif
