/////////////////////////////////////////////////////////////////////////////
//
// File: asm.cpp
//
//
// Assembler for the R1-emulator:
// ------------------------------
//
// The assembler operates on simple statements like "move 10,r20"
//
// The grammar is:
// ---------------
// 
// See "scanner.h" for definition of <ident>, <const> and more...
//
// *** EXPRESSIONS ***
// E0    ::= <E1> { ("+" | "-") <E1> }
// E1    ::= <E2> { "*" <E2> }
// E2    ::= <value> | "(" <E0> ")"
// value ::= <const> | <ident(type==const)> | "@" <ident(type!=const)>
//
// *** INSTRUCTIONS ***
//	movemod    ::= "B" | "W"
//	mne_move   ::= "MOVE" [ "." <movemod> ]
//	mne_bitman ::= "BTT" | "BTR" | "BTS" | "BTC" | "SHL" |
//	               "SHR" | "SAR" | "ROL" | "RCL" | "ROR" | "RCR"
//	mne_jmp    ::= "CALL" | "JMP" | "JE" | "JNE" | "JC"	| "JNC" |
//	               "JS"   | "JNS" | "JA" | "JAE" | "JB"	| "JBE" |
//	               "JG"   | "JGE" | "JL"	| "JLE"
//	mne_2reg   ::= "ADD" | "SUB" | "AND" | "OR"	| "XOR" | "TEST"
//	mne_1reg	  ::= "CLR" | "SET" | "PUSH" | "POP" | "INC" | "DEC" | "NOT" | "NEG"
//	mne_0reg	  ::= "HALT" | "RET" | "IRET"
//
//	op0reg     ::= <mne_0reg>
//	op1reg     ::= <mne_1reg> <reg>
//	op2reg     ::= <mne_2reg> <reg> "," <reg>
//	opbitman   ::= <mne_bitman> <reg> "," <reg> | <mne_bitman> <E0> "," <reg>
//	opjmp      ::= <mne_jmp> <reg> | <mne_jmp> <ident(type==label)>
// moveoperand::= <const> | "@" <ident> | "(" E0 ")"  | "[" <reg> "]" | "[" <E0> "]" | <ident> [ "[" <E0> "]" ] | <reg> | <reg> ":" <moveoperand>
//	opmove     ::= <mne_move> <moveoperand> "," <moveoperand>
//
//	inst       ::= <opmove> | <opjmp> | <opbitman> | <op2reg> | <op1reg> | <op0reg>
//
// *** LABELS ***
//	storage_spec ::= "db" | "dw"
//	num_list	 ::= <E0> { "," <E0> }
//	label			::= <ident> "=" <E0 resolvable> | <ident> ":" [ <stor_spec> [ "*" <E2 resolvable> | <num_list> ] | <inst> ]
//
// *** DIRECTIVE ***
// directive   ::= "end" | "assume" <assume_list> | "seg" <ident>
// assume_list ::={ <reg> ":" <ident> }
//
// *** ASSEMBLER LINES ***
//	comment		::= "#"<rest-of-line>
//	line			 ::= [ <label> | <inst> ] [ <comment> ] | "." <directive>
//
#include <string>
#include "asm.h"
#include "error.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <strstream>
#include "disasm.h"
#define DEBUG(s) //(clog << s << endl << flush)


string lalign(int isize, string i, char fill=' ') {
	string res;
	int lp1;
	for (lp1=0; lp1<isize; lp1++)
		res += (lp1<i.length()?i[lp1]:fill);
	return res;
}


// groups:
// E_ - general
// EL_ - label errors
// E2_ - two-reg errors
// E1_ - one-reg errors
// E0_ - zero-reg errors
// EB_ - bitman errors
// EJ_ - jump errors
// EM_ - move errors
// ED_ - directive errors
#define E_UNKNOWNTOKEN 1
#define E_ADDITEMS 2
#define E_NOTANINST 4
#define E_REGEXPECTED 5
#define E_INVALIDREG 6
#define E_UNDECLARED 7
#define E_CONSTEXPECTED 8
#define E_IDENTEXPECTED 9
#define E_CANTTAKEADDR 10
#define E_VALUEEXPECTED 11
#define E_MISSINGPAR 12
#define EL_DUPLICATEIDENT 100
#define EL_DISPLACEMENTERROR 101
#define EL_COUNTEXPECTED 102
#define EL_VALLISTCOMMASEP 103
#define EL_UNPARSED 104
#define E2_INV 200
#define E2_COMMAEXPECTED 201
#define E1_INV 300
#define E0_INV 400
#define EB_INV 500
#define EB_COMMAEXPECTED 501
#define EB_OUTOFBOUNDS 502
#define EJ_INV 600
#define EJ_NODESTINATION 601
#define EJ_INTERSEGMENT 602
#define EJ_NOTLABEL 603

#define EM_INV                  700
#define EM_INVMOVEMOD           701
#define EM_LOPBOTHREGVAL        702
#define EM_CANNOTSTOREIMM       703
#define EM_ROPREGANDBASENOTBP   704
#define EM_NOINSTRUCTION        705
#define EM_ROPERROR             706
#define EM_ROPERROR1            707
#define EM_ROPERROR2            708
#define EM_ROPERROR3            709
#define EM_DOUBLEBASE           710
#define EM_BASEOFFSETEXPECTED   711
#define EM_OFFSETMISSINGBRACKET 712
#define EM_NOASSUME             713
#define EM_ARRAYMISSINGBRACKET  714
#define EM_MISSINGPAR           715
#define EM_NOOPERAND            716

#define EM_COMMABETWEENOP       717
#define EM_LABELASOPERAND       718
#define EM_LOADIMMWITHADD       719

#define ED_INV 800
#define ED_SEGNAMEEXPECTED 801
#define ED_SEGNAMEDUPLICATE 802
#define ED_IDENTEXPECTED 803
#define ED_COLONBETWEENASSUME 804
#define ED_UNKNOWN 805


class ESyntax: public Error {
public:

	ESyntax(int se): Error(se) {}

	ESyntax(int se, string message ): Error(se) {}


	string getMsg() {
		ostrstream msg;
		msg << "STX" << setw(3) << setfill('0') << ecode << ": ";

		switch (ecode) {
			case E_UNKNOWNTOKEN: msg << "Unknown token"; break;
			case E_ADDITEMS: msg << "additional items on line, end-of-line expected"; break;
			case E_NOTANINST: msg << "not an instruction"; break;
			case E_REGEXPECTED: msg << "register expected"; break;
			case E_INVALIDREG: msg << "invalid register!"; break;
			case E_UNDECLARED: msg << "undeclared identifier!"; break;
			case E_CONSTEXPECTED: msg << "identifier must be constant!"; break;
			case E_IDENTEXPECTED: msg << "identifier expected"; break;
			case E_CANTTAKEADDR: msg << "cannot take address of constant"; break;
			case E_VALUEEXPECTED: msg << "value expected"; break;
			case E_MISSINGPAR: msg << "missing )"; break;
			
			case EL_DUPLICATEIDENT: msg << "duplicate identifier"; break;
			case EL_DISPLACEMENTERROR: msg << "label address mismatch in second pass!"; break;
			case EL_COUNTEXPECTED: msg << "* must be followed by a count"; break;
			case EL_VALLISTCOMMASEP: msg << "list of values must be seperated by ,"; break;
			case EL_UNPARSED: msg << "identifier must be followed by either : or ="; break;
			
			case E2_INV: msg << "trying to parse as two-reg when it is not"; break;
			case E2_COMMAEXPECTED: msg << "expected , between operands"; break;
			
			case E1_INV: msg << "trying to parse as one-reg when it is not"; break;
			
			case E0_INV: msg << "trying to parse as zero-reg when it is not"; break;
			
			case EB_INV: msg << "trying to parse as bitman when it is not"; break;
			case EB_COMMAEXPECTED: msg << "expected , between operands"; break;
			case EB_OUTOFBOUNDS: msg << "does'nt it seem a bit stupid to access bits outside range"; break;
			
			case EJ_INV: msg << "trying to parse as jump when it is not"; break;
			case EJ_NODESTINATION: msg << "where to? no destination found"; break;
			case EJ_NOTLABEL: msg << "identifier not defined as label"; break;
			case EJ_INTERSEGMENT: msg << "inter-segment jumps not supported"; break;
			
			case EM_INV: msg << "trying to parse as move when it is not"; break;
			case EM_INVMOVEMOD: msg << "invalid move modifier"; break;
			case EM_LOPBOTHREGVAL: msg << "left operand is not offset but has both register and value"; break;
			case EM_CANNOTSTOREIMM: msg << "cannot store to immediate"; break;
			case EM_ROPREGANDBASENOTBP: msg << "right operand is register and has base different from BP"; break;
			case EM_NOINSTRUCTION: msg << "no instruction matches operands"; break;
			case EM_ROPERROR: msg << "error in right operand"; break;
			case EM_ROPERROR1: msg << "right operand is not offset but has value, base or no register"; break;
			case EM_ROPERROR2: msg << "right operand is offset but has value or a base-reg different from BP"; break;
			case EM_ROPERROR3: msg << "right operand must be either be register or register-as-offset"; break;
			case EM_DOUBLEBASE: msg << "only one base-register per operand allowed"; break;
			case EM_BASEOFFSETEXPECTED: msg << "offset expected after base"; break;
			case EM_OFFSETMISSINGBRACKET: msg << "missing ] after offset"; break;
			case EM_NOASSUME: msg << "no assumptions could resolve the base for identifier"; break;
			case EM_ARRAYMISSINGBRACKET: msg << "array-indexing missing ]"; break;
			case EM_MISSINGPAR: msg << "missing ) after expression"; break;
			case EM_NOOPERAND: msg << "missing operand!"; break;
			case EM_COMMABETWEENOP: msg << ", expected between operand"; break;
			case EM_LABELASOPERAND: msg << "label not allowed as operand"; break;
			case EM_LOADIMMWITHADD: msg << "can only load immediate into register"; break;

			case ED_INV: msg << "trying to parse directive that does not start with ."; break;
			case ED_SEGNAMEEXPECTED: msg << "segment name expected after .seg"; break;
			case ED_SEGNAMEDUPLICATE: msg << "segment name must be uniqe in file"; break;
			case ED_IDENTEXPECTED: msg << "identifier expected"; break;
			case ED_COLONBETWEENASSUME: msg << "must seperate assumes with :"; break;
			case ED_UNKNOWN: msg << "unknown assembler directive"; break;
			default: msg << "???";
		}
		if (emsg!="") msg << " - " << emsg;
		msg << ends;
		return msg.str();
	}
};



//////////////////////////////////////////////////////////////////////////////
//
// Identifiers:
//
class Ident {
	public:
		// type of this identifier
		enum IdentType { ID_CONST, ID_LABEL, ID_BYTE, ID_WORD };
		IdentType type;
		// name of this identifier
		string name;
		// address of this identifier
		WORD addr;
		// segment in which this identifier is located
		pSeg segm;

		Ident(): name(""), addr(0), type(ID_CONST), segm(NULL) {}

		Ident( string nname, IdentType ttype, WORD aaddr=0, pSeg ssegm=NULL ):
			name(nname), addr(aaddr), type(ttype), segm(ssegm) {}
};



//////////////////////////////////////////////////////////////////////////////
//
// Segment:
//
// Segments are used to implement the different storage-areas in a file.
// Each segment has a name, a counter and a lot of data.
class Segment {
	public:
		Segment(string);
		virtual ~Segment();

		// returns the current index of the segment
		WORD getCount() { return count; }

		// returns name of segment
		string getName() { return name; }

		// increases segment size by count. Cannot be done when the segment has
		// been realised.
		WORD inc( WORD=1 );

		// realize segment. Only when the segment is realized can data be added
		// to the segment. If realisation fails, an exception may be thrown.
		void realize();

		// sets data-area. Throws exception if not realized.
		void set( WORD, BYTE );
		void set( WORD, WORD );
		void put( BYTE );
		void put( WORD );

		// Will write segment to stream in format:
		// struct segment {
		//	 BYTE segname[32];
		//	 WORD seglength;
		//	 BYTE segdata[seglength];
		// }
		void writeToStream(ostream& out);

		// use disassembler to disassemble segment
		void view( ostream& o );

	private:
		string name;
		WORD count;
		BYTE* data;
		WORD size;
};


Segment::Segment(string nname): name(nname), count(0), data(NULL) {}

Segment::~Segment() {
	if (data) delete data;
}

WORD Segment::inc( WORD size=1 ) {
	WORD oldcount = count;
	count += size;
	return oldcount;
}

void Segment::realize() {
	if (data) throw Error(100,"Segment::realize() - already realised");
	data = new BYTE[count];
	size = count;
	// reset memory area
	BYTE *d=data;
	for (int lp1=0; lp1<size; lp1++, d++)
		*d = 0;
	count = 0;
}


void Segment::set( WORD addr, BYTE dat ) {
	if (!data) throw Error(101, "Out of bounds in segment:	"+name);
	if (addr>size) throw Error(101, "Out of bounds in segment: "+name);
	*(data+addr)=dat;
}

void Segment::set( WORD addr, WORD dat ) {
	if (!data) throw Error(100, "Segment not realized: "+name);
	if (addr+1>size) throw Error(101, "Out of bounds in segment: "+name);
	*(data+addr+0)=dat>>16;
	*(data+addr+1)=dat & 0x0ffff;
}



void Segment::put( BYTE b ) {
	if (data) *(data+count) = b;
	count++;
}

void Segment::put( WORD w ) {
	if (data) *(data+count) = w >> 16;
	count++;
	if (data) *(data+count) = w & 0x0ffff;
	count++;
}


void Segment::writeToStream(ostream& out) {
	SegHeader segh;

	for (int lp1=0; lp1<32; lp1++)
		 segh.name[lp1] = (lp1<name.length()? name[lp1] : 0);
	segh.length = size;
	out.write(&segh,sizeof(segh));
	out.write(data,segh.length*sizeof(BYTE));
}


void Segment::view( ostream& o) {
	DisAsm d;
	d.disAsmMem(data, size, o);
}

//////////////////////////////////////////////////////////////////////////////
// ASSEMBLER
//

// The constructor for the assembler will initialize all fields
Assembler::Assembler() {
	segments["code"] = new Segment("code");
	segments["data"] = new Segment("data");
	curseg = segments["code"];
	passCount = psFirst;
}


// clear all segments and idents.
Assembler::~Assembler() {
	map<string, pSeg>::iterator segit = segments.begin();
	while (segit!=segments.end()) {
		delete segit->second;
		segit++;
	}
	segments.clear();
	idents.clear();
}

// Move to the next token in the input
void Assembler::nextToken() {
	curtok = sc.next();
}


// Look-up identifier in identifier base
Ident Assembler::lookup(string id) {
	map<string,Ident>::iterator l = idents.find(id);
	if (l==idents.end())
		throw ESyntax(7, id);
	return l->second;
}


// Check for identifier in both segment- and ident-lists!
bool Assembler::idExists(string id) {

	map<string,Ident>::iterator l = idents.find(id);
	if (l!=idents.end())
		return true;

	map<string,pSeg>::iterator s = segments.find(id);
	if (s!=segments.end())
		return true;

	return false;
}


int Assembler::assemble( string infile, string outfile, string mapfile ) {
	char buf[256];
	ifstream in(infile.c_str());
	bool errors;


	// first pass
	cout << "R1-assembler (c) 2000 MeeSoft (GPL)" << endl;
	cout << "-----------------------------------" << endl;
	cout << "Source:	 " << infile << endl;
	cout << "Object:	 " << outfile << endl;
	cout << "Map file: " << mapfile << endl;
	cout << endl;

	cout << "First pass..." << endl;

	eof = false;
	errors = false;
	lineno = 0;
	for (int lp1=0; lp1<32; lp1++) assumes[lp1] = NULL;
	while (!in.eof() && !eof) try {
			in.getline(buf,256);
			lineno++;
			prsLine(buf);
		} catch (ESyntax& e) {
			errors=true;
			cerr << infile << "(" << dec << lineno << "): " << e.getMsg() << endl;
		} catch (Error& e) {
			cerr << infile << "(" << dec << lineno << "): FATAL: " << e.getMsg() << endl;
			return -1;
		}
	if (errors) return -1;

	cout << "No syntax errors detected in first pass!" << endl << endl;

	// realize all segments
	cout << "Realizing segments in memory...";
	for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++)
		 p->second->realize();
	cout << "done" << endl << endl;

	// second pass
	ofstream fmapfile;
	bool makemapfile = mapfile!="";
	if (makemapfile) {
		fmapfile.open(mapfile.c_str());
		fmapfile << "FILE:	 " << mapfile << endl;
		fmapfile << "SOURCE: " << infile << endl;
		fmapfile << "OBJECT: " << outfile << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Source-listing and segment counter: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
	}

	cout << "Second pass - final assembly..." << endl;
	in.seekg(0);
	lineno = 0;
	passCount = psFinal;
	eof = false;
	for (int lp1=0; lp1<32; lp1++) assumes[lp1] = NULL;
	try {
		while (!in.eof() && !eof) {
			in.getline(buf,256);
			lineno++;
			if (makemapfile) fmapfile << curseg->getName() << ":$" << hex << setw(4) << setfill('0') << curseg->getCount() << ">" << buf << endl;
			prsLine(buf);
		}
	} catch (ESyntax& e) {
			cerr << infile << "(" << dec << lineno << "): " << e.getMsg() << endl;
		return -1;
	}
	in.close();
	cout << endl;

	cout << "Writing object file...";
	fstream out(outfile.c_str(), ios::out | ios::binary );
	for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++)
		 p->second->writeToStream(out);
	out.close();
	cout << "done" << endl << endl;

	if (makemapfile) {
		fmapfile << endl;

		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Identifiers: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;

		fmapfile << lalign(20,"Name:") << " ";
		fmapfile << lalign(10,"Segment:") << " ";
		fmapfile << lalign(11,"Address:") << " ";
		fmapfile << lalign( 5,"Type:") << endl << endl;

		string idtype[] = { "const","label","byte","word"};
		for (map<string, Ident>::iterator i=idents.begin(); i!=idents.end(); i++) {
			fmapfile << lalign(20,i->second.name) << " ";
			fmapfile << lalign(10,(i->second.segm!=NULL?i->second.segm->getName():string("-"))) << " ";
			fmapfile << setw( 8) << hex << setfill('0') << i->second.addr << setfill(' ') << "h	 ";
			fmapfile << lalign(5,idtype[i->second.type]) << endl;
		}
		fmapfile << endl;

		fmapfile << lalign(80,"",'-') << endl << endl;
		fmapfile << "* Segments: " << endl << endl;
		fmapfile << lalign(80,"",'-') << endl << endl;
		for (map<string,pSeg>::iterator p=segments.begin(); p!=segments.end(); p++) {
			fmapfile << "Segment: " << p->second->getName() << endl << endl;
			p->second->view(fmapfile);
			fmapfile << endl;
		}
		fmapfile << endl << "----==== EOF ====----" << endl;
		fmapfile.close();
	}

	cout << "Done assembling..." << endl;
	return 0;
}


// Assemble assembles an instruction from a statement.
void Assembler::prsLine( string line ) {
	sc.newLine(line);
	nextToken();

	// empty line
	switch (curtok) {
		case TT_EOL: // empty line
			return;
			break;
			
		case TT_DOT: // directive
			prsDirective();
			break;
			
		case TT_ID:  // label
			prsLabel();
			if (curtok!=TT_EOL) prsInst();
			break;
			
		default: // instruction
			prsInst();
	}
			
}


// <directive> ::= "." <dir>
void Assembler::prsDirective() {
	DEBUG("Assembler::prsDirective()");

	if (curtok!=TT_DOT) throw ESyntax(ED_INV);
	nextToken();

	switch (curtok) {
		case TT_SEG: {// "." "seg" <id>
			for (int lp1=0; lp1<32; lp1++) assumes[lp1]=NULL;
			nextToken();
			if (curtok!=TT_ID) throw ESyntax(ED_IDENTEXPECTED);
			string id = sc.getLex();
			if (idents.find(id)!=idents.end()) throw ESyntax(ED_SEGNAMEDUPLICATE, id);
			if (segments[id]==NULL) segments[id] = new Segment(id);
			curseg=segments[id];
			break;
		}
			
		case TT_ASSUME: { // "." "assume" { <reg>":"<ident> }
			for (int lp1=0; lp1<32; lp1++) assumes[lp1]=NULL;
			nextToken();
			while (curtok==TT_REG) {
				int regno = prsReg();
				if (curtok!=TT_COLON) throw ESyntax(ED_COLONBETWEENASSUME);
				nextToken();
				if (curtok!=TT_ID) throw ESyntax(ED_SEGNAMEEXPECTED);
				string id = sc.getLex();
				nextToken();
				// look for segment in table - verify the existance of a
				// segment of that name. If a segment is found, add the pair
				// to the assumptions, so the other may look it up and find
				// the register/segment combination.
				if (passCount==psFinal) {
					if (segments.find(id)==segments.end()) throw ESyntax(ED_SEGNAMEEXPECTED, id);
					assumes[regno]=segments[id];
				}
			}
			break;
		}
			
		case TT_END: 
			eof = true;
			break;
			
		default:
			throw ESyntax(ED_UNKNOWN);
	}
}




// parses an expanded value-expression
// <E0> ::= <E1> { ("+"|"-") <E1> }
INT Assembler::prsExpr0( bool resolve ) {
	DEBUG("Assembler::prsExpr0()");
	WORD v = prsExpr1( resolve );
	while (curtok==TT_PLUS || curtok==TT_MINUS) {

		if (curtok==TT_PLUS) {
			nextToken();
			v += prsExpr1( resolve );
		}
		if (curtok==TT_MINUS) {
			nextToken();
			v -= prsExpr1( resolve );
		}
	}
	return v;
}


// parses an expanded value-expression
// <E1> ::= <E2> { "*" <E2> }
INT Assembler::prsExpr1( bool resolve ) {
	DEBUG("Assembler::prsExpr1()");
	WORD v = prsExpr2( resolve );
	while (curtok==TT_MULT) {
		nextToken();
		v *= prsExpr2( resolve );
	}
	return v;
}


// parses an expanded value-expression
// <E2> ::= "(" <E0> ")" | <value>
INT Assembler::prsExpr2( bool resolve ) {
	DEBUG("Assembler::prsExpr2()");
	WORD v;
	if (curtok==TT_LP) {
		nextToken();
		v = prsExpr0( resolve );
		if (curtok!=TT_RP) throw ESyntax(E_MISSINGPAR);
		nextToken();
	} else
		v = prsValue( resolve );
	return v;
}


// This method parses a simple value -
// <value> ::= <const> | <ident(type==const)> | "@" <ident(type!=const)>
INT Assembler::prsValue( bool resolve ) {
	DEBUG("Assembler::prsValue()");
	WORD v;
	switch (curtok) {
		// read constant
		case TT_CONST:
			v = sc.getWord();
			break;
 
		// read identifier (must be constant)
		case TT_ID:
			if (resolve) {
				Ident i = lookup(sc.getLex());
				if (i.type!=Ident::ID_CONST) throw ESyntax(E_CONSTEXPECTED, i.name);
				v = i.addr;
			} else
				v = 0;
			break;
 
		// read absolute address
		case TT_AT:
			nextToken();
			if (curtok!=TT_ID) throw ESyntax(E_IDENTEXPECTED, sc.getLex());
			if (resolve) {
				Ident i = lookup(sc.getLex());
				if (i.type==Ident::ID_CONST) throw ESyntax(E_CANTTAKEADDR, i.name);
				v = i.addr;
			} else
				v = 0;
			break;

		default: throw ESyntax(E_VALUEEXPECTED);
	}
	nextToken();
	return v;
}


// parses a register - returns regno.
int Assembler::prsReg() {
	DEBUG("Assembler::prsReg()");
	if (curtok!=TT_REG) throw ESyntax(E_REGEXPECTED);

	string s = sc.getLex();
	nextToken();
	if (s=="ip") return 0;
	if (s=="flags") return 1;
	if (s=="sp") return 2;
	if (s=="bp") return 3;
	if (s=="it") return 4;
	// read Rxx
	int regno=100;
	if (s[0]=='r' && s.length()==3)
		regno = (s[1]-'0')*10+(s[2]-'0');
	if (regno>31) throw ESyntax(6);
	return regno;
}


// <label> ::= <id> "=" <number> | <id> ":" [<ss> ["*" <number> | { <value> "," } ] ]
void Assembler::prsLabel() {
	string id = sc.getLex();
	DEBUG("Assembler::prsLabel() " << id );
	
	// avoid duplicate identifiers
	if (passCount==psFirst && idExists(id) ) throw ESyntax(EL_DUPLICATEIDENT, id);
	nextToken();
	
	// constant - must read number after
	if (curtok==TT_EQ) {
		DEBUG(" - found constant");
		nextToken();
		WORD val = prsExpr0( true ); // must resolve to immediate value!
		// add constant to idents if first-pass.
		idents[id] = Ident(id, Ident::ID_CONST, val);

		if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
		return;
	}
	// label - expects instruction or storagespecifier
	if (curtok==TT_COLON) {
		DEBUG(" - found label: " << id);

		WORD adr = curseg->getCount();

		if (passCount!=psFinal) idents[id] = Ident(id, Ident::ID_LABEL, adr, curseg );
		if (passCount==psFinal) if (adr!=idents[id].addr) throw ESyntax(EL_DISPLACEMENTERROR, id);
		nextToken();
		// catch storage-specifiers
		if (curtok==TT_DW || curtok==TT_DB) {
			DEBUG("	 - found storage");
			int itmsize = 0;

			if (curtok==TT_DB) {
				if (passCount!=psFinal) idents[id].type=Ident::ID_BYTE;
				itmsize = 1;
			}
			if (curtok==TT_DW) {
				if (passCount!=psFinal) idents[id].type=Ident::ID_WORD;
				itmsize = 2;
			}

			nextToken();
			// look for "*" <E2 that is resolvable>
			if (curtok==TT_MULT) {
				nextToken();
				INT count = prsExpr2(true);
				curseg->inc(itmsize*count);
				nextToken();
				if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
				return;
			}
			// look for value list (more on line)
			if (curtok!=TT_EOL) {
				WORD offset = 0;
				while (curtok!=TT_EOL) {
					if (curtok==TT_STRING) {
						string s = sc.getLex();
						DEBUG("In vallist: found the string: " << s );
						for (int lp1=0; lp1<s.length(); lp1++)
							if (itmsize==1) curseg->put( static_cast<BYTE>(s[lp1]) );
							else curseg->put( static_cast<WORD>(s[lp1]) );
						nextToken();
					} else {
						WORD val = prsExpr0(passCount==psFinal);
						if (itmsize==1) curseg->put( static_cast<BYTE>(val) );
						else curseg->put( static_cast<WORD>(val) );
					}
					if (curtok!=TT_COMMA && curtok!=TT_EOL) throw ESyntax(EL_VALLISTCOMMASEP);
					if (curtok==TT_COMMA) nextToken();
				}
				if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
				return;
			}
			// if neither value-list or rep-count use default one instance uninitialised
			curseg->inc(itmsize);
			return ;
		}
		// a storage-specifier did not follow the label, so an instruction may
		// be here
		DEBUG(" - no storage-specifier, crude label");
		return ;
	} 
	throw ESyntax(EL_UNPARSED);
}






// prsInst parses an instruction according to:
//
// <inst> ::= <opmove> <regofsval> "," <regofsval> |
//						<opjmp> <regofs> |
//						<opbitman> <regval> |
//						<op2> <reg> "," <reg> |
//						<op1> <reg> |
//						<op0>
void Assembler::prsInst() {
	DEBUG("Assembler::prsInst()");
	switch (curtok) {
		case TT_JMP: case TT_CALL:
		case TT_JE: case TT_JNE: case TT_JC: case TT_JNC: case TT_JS: case TT_JNS:
		case TT_JA: case TT_JAE: case TT_JB: case TT_JBE: case TT_JG: case TT_JGE: case TT_JL: case TT_JLE:
			prsJmp();
			break;

		case TT_SHR: case TT_SAR: case TT_SHL:
		case TT_ROR: case TT_RCR: case TT_ROL: case TT_RCL:
		case TT_BTT: case TT_BTR: case TT_BTC: case TT_BTS:
			prsBitman();
			break;

		case TT_MOVE:
			prsMove();
			break;


		case TT_ADD: case	TT_ADC:  case TT_SUB: case TT_SBB: case TT_CMP:
		case TT_AND: case TT_TEST: case TT_OR:  case TT_XOR:
		case TT_MUL: case TT_IMUL: case TT_DIV: case TT_IDIV:
			prs2reg();
			break;

		case TT_CLR:  case TT_SET:
		case TT_NOT:  case TT_NEG:
		case TT_INC:  case TT_DEC:
		case TT_PUSH: case TT_POP:
			prs1reg();
			break;

		case TT_NOP: case TT_RET: case TT_IRET: case TT_HALT:
			prs0reg();
			break;

		default: throw ESyntax(E_NOTANINST);
	}
}


// Parses two-register instructions
// Puts result in current seg on counter location if psFinal
// <op2reg> ::= <op> <reg> "," <reg>
void Assembler::prs2reg() {
	DEBUG("Assembler::prs2reg()");
	Opcode o;
	o.opcode = 0;
	switch (curtok) {
		case TT_ADD:  o.twoReg.inst = OC_ADD; o.twoReg.param1 = 0; o.twoReg.param2 = 0; break;
		case TT_ADC:  o.twoReg.inst = OC_ADD; o.twoReg.param1 = 0; o.twoReg.param2 = 1; break;
		case TT_SUB:  o.twoReg.inst = OC_SUB; o.twoReg.param1 = 0; o.twoReg.param2 = 0; break;
		case TT_SBB:  o.twoReg.inst = OC_SUB; o.twoReg.param1 = 0; o.twoReg.param2 = 1; break;
		case TT_CMP:  o.twoReg.inst = OC_SUB; o.twoReg.param1 = 1; o.twoReg.param2 = 0; break;
		case TT_AND:  o.twoReg.inst = OC_LOGICAL; o.twoReg.param1 = 0; o.twoReg.param2 = 0; break;
		case TT_TEST: o.twoReg.inst = OC_LOGICAL; o.twoReg.param1 = 0; o.twoReg.param2 = 1; break;
		case TT_OR:   o.twoReg.inst = OC_LOGICAL; o.twoReg.param1 = 1; o.twoReg.param2 = 0; break;
		case TT_XOR:  o.twoReg.inst = OC_LOGICAL; o.twoReg.param1 = 1; o.twoReg.param2 = 1; break;
		case TT_MUL:  o.twoReg.inst = OC_MULDIV; o.twoReg.param1 = 0; o.twoReg.param2 = 0; break;
		case TT_IMUL: o.twoReg.inst = OC_MULDIV; o.twoReg.param1 = 0; o.twoReg.param2 = 1; break;
		case TT_DIV:  o.twoReg.inst = OC_MULDIV; o.twoReg.param1 = 1; o.twoReg.param2 = 0; break;
		case TT_IDIV: o.twoReg.inst = OC_MULDIV; o.twoReg.param1 = 1; o.twoReg.param2 = 1; break;
		default: throw ESyntax(E2_INV);
	}
	nextToken();
	o.twoReg.source = prsReg();
	if (curtok!=TT_COMMA) throw ESyntax(E2_COMMAEXPECTED);
	nextToken();
	o.twoReg.dest = prsReg();

	curseg->put(o.opcode);

	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// Parses one-register instructions
// Puts result in current seg on counter location if psFinal
void Assembler::prs1reg() {
	DEBUG("Assembler::prs1reg()");
	Opcode o;
	o.opcode = 0x0ffff;
	switch (curtok) {
		case TT_CLR:  o.oneReg.inst = OC_CLR; break;
		case TT_SET:  o.oneReg.inst = OC_SET; break;
		case TT_NOT:  o.oneReg.inst = OC_NOT; break;
		case TT_NEG:  o.oneReg.inst = OC_NEG; break;
		case TT_INC:  o.oneReg.inst = OC_INC; break;
		case TT_DEC:  o.oneReg.inst = OC_DEC; break;
		case TT_PUSH: o.oneReg.inst = OC_PUSH; break;
		case TT_POP:  o.oneReg.inst = OC_POP; break;
		default: throw ESyntax(E1_INV);
	}
	nextToken();
	o.oneReg.reg = prsReg();

	curseg->put(o.opcode);

	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// Parses zero-register instructions
// Puts result in current seg on counter location if psFinal
void Assembler::prs0reg() {
	DEBUG("Assembler::prs0reg()");
	Opcode o;
	o.opcode = 0x0ffff;
	switch (curtok) {
		case TT_NOP:  o.opcode = 0; break;
		case TT_RET:  o.zeroReg.inst = OC_RET; break;
		case TT_IRET: o.zeroReg.inst= OC_IRET; break;
		case TT_HALT: o.zeroReg.inst = OC_HALT; break;
		default: throw ESyntax(E0_INV);
	}
	nextToken();
	curseg->put(o.opcode);
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}


// <opjmp> ::= <mne_jmp> <label> | <mne_jmp> <reg>
void Assembler::prsJmp() {
	DEBUG("Assembler::prsJmp()");
	int oc;
	Opcode o;

	switch (curtok) {
		case TT_JMP:  oc = OC_JMP;  break;
		case TT_CALL: oc = OC_CALL; break;
		case TT_JE:   oc = OC_JE;   break;
		case TT_JNE:  oc = OC_JNE;  break;
		case TT_JC:   oc = OC_JC;   break;
		case TT_JNC:  oc = OC_JNC;  break;
		case TT_JS:   oc = OC_JS;   break;
		case TT_JNS:  oc = OC_JNS;  break;
		case TT_JA:   oc = OC_JA;   break;
		case TT_JAE:  oc = OC_JAE;  break;
		case TT_JB:   oc = OC_JB;   break;
		case TT_JBE:  oc = OC_JBE;  break;
		case TT_JG:   oc = OC_JG;   break;
		case TT_JGE:  oc = OC_JGE;  break;
		case TT_JL:   oc = OC_JL;   break;
		case TT_JLE:  oc = OC_JLE;  break;
		default: throw ESyntax(EJ_INV);
	}
	nextToken();
	switch (curtok) {
		case TT_REG: // register - resolveable at first
			o.oneReg.reserved=0xffff;
			o.oneReg.inst = oc;
			o.oneReg.reg = prsReg();
			curseg->put(o.opcode);
			break;
			
		case TT_ID: // identifier - if first-pass, assume 00000000
			o.zeroReg.reserved=0xffff;
			o.zeroReg.inst=oc;
			curseg->put(o.opcode);
			if (passCount==psFinal) { // look-up label. Control segments.
				Ident i = lookup(sc.getLex());
				if (i.type!=Ident::ID_LABEL) throw ESyntax(EJ_NOTLABEL, i.name);
				if (i.segm!=curseg) throw ESyntax(EJ_INTERSEGMENT, i.segm->getName()+":"+i.name );
				curseg->put( i.addr-(curseg->getCount()+2) ); // corrected +2 due to instruction pointer moving
			} else {
				curseg->put((WORD)0);
			}
			break;
			
		default: // syntax error - no destination
			throw ESyntax(EJ_NODESTINATION);
	}
	nextToken();
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);
}



// <opbitman> ::= <mnemonic_bitman> <reg>","<reg> | <mnemonic_bitman> <val>","<reg>
void Assembler::prsBitman() {
	DEBUG("Assembler::prsBitman()");
	Opcode o;
	switch (curtok) {
		case TT_BTT: o.twoReg.inst=OC_BITTEST; o.twoReg.param1=0; break;
		case TT_BTR: o.twoReg.inst=OC_BITTEST; o.twoReg.param1=1; break;
		case TT_BTC: o.twoReg.inst=OC_BITMAN; o.twoReg.param1=0; break;
		case TT_BTS: o.twoReg.inst=OC_BITMAN; o.twoReg.param1=1; break;
		case TT_SHR: o.twoReg.inst=OC_SHR; o.twoReg.param1=0; break;
		case TT_SAR: o.twoReg.inst=OC_SHR; o.twoReg.param1=1; break;
		case TT_SHL: o.twoReg.inst=OC_SHL; o.twoReg.param1=0; break;
		case TT_ROR: o.twoReg.inst=OC_ROR; o.twoReg.param1=0; break;
		case TT_RCR: o.twoReg.inst=OC_ROR; o.twoReg.param1=1; break;
		case TT_ROL: o.twoReg.inst=OC_ROL; o.twoReg.param1=0; break;
		case TT_RCL: o.twoReg.inst=OC_ROL; o.twoReg.param1=1; break;
		default: throw ESyntax(EB_INV);
	}
	nextToken();

	if (curtok==TT_REG) {
		o.twoReg.param2=0;
		o.twoReg.source=prsReg();
	} else {
		o.twoReg.param2=1;
		INT v = prsExpr0(passCount==psFinal);
		if (passCount==psFinal && (v<0 || v>31) ) throw ESyntax(EB_OUTOFBOUNDS);
		o.twoReg.source = v;
	}
	if (curtok!=TT_COMMA) throw ESyntax(EB_COMMAEXPECTED);
	nextToken();
	o.twoReg.dest=prsReg();
	if (curtok!=TT_EOL) throw ESyntax(E_ADDITEMS);

	curseg->put(o.opcode);
}



struct moveOp {
	bool bytemove; // set if move is byte
	bool isOfs;		// set if this is offset
	bool hasBase;	// set if this has base-register
	int	baseReg;	// base register no.
	bool hasReg;	 // set if register found
	int	regNo;		// register no
	bool hasVal;	 // set if value found
	WORD val;			// value
};

// <opmove> ::= "move" [ "." "b" | "w" ]
void Assembler::prsMove() {
	DEBUG("Assembler::prsMove()");
	moveOp lop, rop;
	bool bytemove = false;
	bool hasmod = false;
	Opcode o;

	if (curtok!=TT_MOVE) throw ESyntax(EM_INV);
	nextToken();

	// support ".b" and ".w"
	if (curtok==TT_DOT) {
		nextToken();
		hasmod = true;
		if (sc.getLex()=="b") bytemove = true;
		else if (sc.getLex()=="w") bytemove = false;
		else throw ESyntax(EM_INVMOVEMOD, sc.getLex() );
		nextToken();
	}
	// read the two operands
	lop = prsMoveOperand();
	if (curtok!=TT_COMMA) throw ESyntax(EM_COMMABETWEENOP);
	nextToken();
	rop = prsMoveOperand();

	// now, assemble the instruction

	// store-instructions:
	// left-operand is not an offset
	if (!lop.isOfs) {
		// if left-op is not an offset but has both reg and value
		// then syntax-error!
		//<reg>+<val>, ...
		if (lop.hasVal && lop.hasReg) throw ESyntax(EM_LOPBOTHREGVAL);
		if (rop.hasVal && !rop.isOfs) throw ESyntax(EM_CANNOTSTOREIMM);

		// <imm> -> <reg>
		if (lop.hasVal) {
			// left-op is value (without reg) and right-op is reg without val
			// then load-imm
			if (lop.hasReg || lop.hasBase || rop.isOfs || !rop.hasReg || rop.hasVal || rop.hasBase) throw ESyntax(EM_LOADIMMWITHADD);
			o.oneReg.reserved = 0xffff;
			o.oneReg.inst = (bytemove?OC_LOADIMM_H:OC_LOADIMM);
			o.oneReg.reg = rop.regNo;
			curseg->put(o.opcode);
			if (bytemove) curseg->put( (BYTE) lop.val );
			else curseg->put(lop.val);
			return;
		}

		// <reg> -> <reg>
		// <reg> -> "["<reg>"]"
		if (rop.hasReg && !rop.hasVal) {
			if (rop.hasBase && rop.baseReg!=3) throw ESyntax(EM_ROPREGANDBASENOTBP);
			o.twoReg.inst = (bytemove?OC_MOVE_H:OC_MOVE);
			o.twoReg.param1 = 0; o.twoReg.param2 = rop.isOfs;
			o.twoReg.source = lop.regNo;
			o.twoReg.dest = rop.regNo;
			curseg->put(o.opcode);
			return;
		}

		// <reg> -> "["..."]"
		if (rop.isOfs) {
			if (!rop.hasBase) rop.baseReg=3;
			o.twoReg.inst = OC_MOVEBO;
			o.twoReg.param1 = (hasmod?bytemove:rop.bytemove); o.twoReg.param2 = 1;
			o.twoReg.source = lop.regNo;
			o.twoReg.dest = rop.baseReg;
			curseg->put(o.opcode);
			curseg->put(rop.val);
			return;
		}

		throw ESyntax(EM_NOINSTRUCTION);
	}

	if (lop.isOfs) {
		// if right-op is an offset or has value, throw exception
		// then syntax-error!
		// ... , <reg>+<val>
		if (!rop.isOfs && (rop.hasVal || rop.hasBase || !rop.hasReg) ) throw ESyntax(EM_ROPERROR1);
		if (rop.isOfs && (rop.hasVal || (rop.hasBase&&rop.baseReg!=3)) ) throw ESyntax(EM_ROPERROR2);

		// [<reg>] , [<reg>] or [<reg>] -> <reg>
		if (!lop.hasVal && !rop.hasVal) {
			if ( (lop.hasBase && lop.baseReg!=3) || (rop.hasBase && rop.baseReg!=3) ) throw ESyntax(EM_ROPERROR3);
			o.twoReg.inst = (bytemove?OC_MOVE_H:OC_MOVE);
			o.twoReg.param1 = 1; o.twoReg.param2 = rop.isOfs;
			o.twoReg.source = lop.regNo;
			o.twoReg.dest = rop.regNo;
			curseg->put(o.opcode);
			return;
		}

		// <reg>:[ofs] -> <reg>
		if (lop.hasVal && !rop.isOfs && !rop.hasVal) {
			if (!lop.hasBase) lop.baseReg=3;
			o.twoReg.inst = OC_MOVEBO;
			//if (rop.bytemove!=bytemove) throw ESyntax(603);
			o.twoReg.param1 = (hasmod?bytemove:lop.bytemove); o.twoReg.param2 = 0;
			o.twoReg.source = lop.baseReg;
			o.twoReg.dest = rop.regNo;
			curseg->put(o.opcode);
			curseg->put(lop.val);
			return;
		}
		throw ESyntax(EM_NOINSTRUCTION);
	}

	throw ESyntax(EM_NOINSTRUCTION);
}



// moveoperand ::= <const> | "@" <ident> | "(" E0 ")"  |
//                 "[" <reg> "]" | "[" <E0> "]" |
//	                <ident> [ "[" <E0> "]" ] |
//                 <reg> | <reg> ":" <moveoperand> |  
moveOp Assembler::prsMoveOperand(bool mustfindbase=true) {
	DEBUG("Assembler::prsMoveOperand(" << mustfindbase << ")");
	moveOp d;
	memset(&d,0,sizeof(d));

	// <const>
	if (curtok==TT_CONST) {
		d.hasVal = true;
		d.val = prsValue(passCount==psFinal);
		return d;
	}
	
	// "@" <ident>
	if (curtok==TT_AT) {
		d.hasVal = true;
		d.val = prsValue(passCount==psFinal);
		return d;
	}
	
	// "(" <E0> ")"
	if (curtok==TT_LP) {
		nextToken();
		d.hasVal = true;
		d.val = prsExpr0(passCount==psFinal);
		if (curtok!=TT_RP) throw ESyntax(EM_MISSINGPAR);
		nextToken();
		return d;
	}

	// "[" <reg> "]" | "[" <E0> "]"
	if (curtok==TT_LB) {
		nextToken();
		d.isOfs=true;
		if (curtok==TT_REG) {
			d.hasReg=true;
			d.regNo=prsReg();
		} else {
			d.hasVal=true;
			d.val=prsExpr0(passCount==psFinal);
		}
		if (curtok!=TT_RB) throw ESyntax(EM_OFFSETMISSINGBRACKET);
		// force default base-register on
		if (!d.hasReg && mustfindbase) {
			d.hasReg=true;
			d.regNo=3;
		}
		nextToken();
		return d;
	}

	// <id> [ "[" <val> "]" ]
	if (curtok==TT_ID) {
		string id = sc.getLex();
		nextToken();
		if (passCount==psFinal) {
			Ident i = lookup(id);
			d.hasVal = true;
			d.val = i.addr;
			if (i.type==Ident::ID_CONST) {
				d.isOfs = false;
				d.hasReg = false;
				return d;
			}
			if (i.type==Ident::ID_LABEL) throw ESyntax(EM_LABELASOPERAND, i.name);
			if (i.type==Ident::ID_BYTE) d.bytemove = true;
			if (i.type==Ident::ID_WORD) d.bytemove = false;
			d.isOfs = true;
			// select correct register based on assume (if required)
			if (mustfindbase) {
				for (int lp1=0; lp1<32; lp1++)
					if (assumes[lp1]==i.segm) {
						d.baseReg = lp1;
						d.hasBase = true;
						break;
					}
				// if current-segment use IP+relative
				if (i.segm==curseg) {
					d.baseReg = 0;
					d.hasBase = true;
					d.val = d.val - (curseg->getCount()+3);
				}
				if (!d.hasBase) throw ESyntax(EM_NOASSUME, id);
			}
		} else {
			d.isOfs = true;
			d.hasBase = mustfindbase;
			d.hasVal = true;
		}
		if (curtok==TT_LB) {
			nextToken();
			WORD eofs = prsExpr0(passCount==psFinal);
			if (!d.bytemove) eofs *= 2;
			d.val += eofs;
			if (curtok!=TT_RB) throw ESyntax(EM_ARRAYMISSINGBRACKET);
			nextToken();
		}
		return d;
	}

	// <reg> [ ":" <regofsval> ]
	if (curtok==TT_REG) {
		d.hasReg=true;
		d.regNo=prsReg();
		if (curtok==TT_COLON) {
			nextToken();
			d.hasReg=false; d.hasBase=true;
			d.baseReg=d.regNo; d.regNo=0;
			moveOp e = prsMoveOperand(false);

			if (e.hasBase) throw ESyntax(EM_DOUBLEBASE);
			if (!e.isOfs) throw ESyntax(EM_BASEOFFSETEXPECTED);
			d.isOfs = e.isOfs;
			d.bytemove = e.bytemove;
			d.hasVal = e.hasVal; d.val = e.val;
			d.hasReg = e.hasReg; d.regNo = e.regNo;
		}
		return d;
	}

	throw ESyntax(EM_NOOPERAND);
}



