#ifndef ASM_H
#define ASM_H

// The Assembler class is a two-pass assembler for the R.1 CPU.
//
// It features segments of sizes up to 32 KBYTE (64 KB Intel)
//
// It produces object-files of the following layout:
//
// 	[Chunk][Chunk][...<eof>
//
//		struct Chunk {
//			 char SegName[16];
//			 unsigned long SegSize;
//			 unsigned char data[SegSize];
//		}
//
//		case SegName of DATA or CODE, a binary image is in the chunk.
//		if SegName="IDEN", an identifier list is present.
//
#include "r1emu.h"
#include <string>
#include <map>
#include "scanner.h"

// This declares the segment-structures
class Segment;
typedef Segment* pSeg;
struct SegHeader {
	BYTE name[32];
	WORD length;
};


struct moveOp;

class Ident;

class Assembler {

	public:
		Assembler();
		~Assembler();


		// run_asm produces an object file from an assembler-file.
		// The object file may later be relocated by adding a stub.
		int assemble( string, string, string );

	private:
		// Scanner used to read file as tokens
		Scanner sc;

		// Current line-number
		int lineno;

		// Segments. Default initialised with "code" and "data"
		// curseg set to "code"
		map<string, pSeg> segments;
		pSeg curseg;

		// Identifiers
		map<string, Ident> idents;

		// assumes: indexes represent registers. The segment pointed to
		// is one of the assumes.
		pSeg assumes[32];


		// Pass-count for assembler. Don't generate code in first.
		enum { psFirst, psFinal } passCount;

		// current token in scanner
		ScTokens curtok;

		// End-of-file
		bool eof;

	public:
		// move to the next token in the input
		void nextToken();

		// Check for identifier in both segment- and ident-lists!
		bool idExists(string id);

		// look-up identifier in database
		Ident lookup(string id);


		// *** EXPRESSIONS ***
		// If resolve=true, a value must be determined during parsing!
		// parses expressions of <E0> ::= <E1> { ("+" | "-") <E1> }
		INT prsExpr0(bool resolve=false);
		
		// parses expressions of <E1> ::= <E2> { "*" <E2> }
		INT prsExpr1(bool resolve=false);
		
		// parses expressions of <E2> ::= <value> | "(" <E0> ")"
		INT prsExpr2(bool resolve=false);

		// <value> ::= <const> | <ident(type==const)> | "@" <ident(type!=const)>
		INT prsValue(bool resolve=false);


		// *** LINES ***
		
		// prsLine will parse a single code-line.
		// If an instruction is present, it will be parsed
		// and a pointer is returned for the assembler to handle
		// prsLine has the following grammer:
		//
		// <line> ::= <directive> | <label> [<instruction> | <storage> ] | <instruction>
		// <directive> ::= "." ( "code" | "data" | "end" )
		// <label> ::= <id> ":"
		// <instruction> ...
		void prsLine( string );

		// "." <directive>
		void prsDirective();

		// prsLine will parse a label
		// <label> ::= <id> ":" [(<spByte>|<spWord>) [ "*" <number> | <value-list> ] ]
		void prsLabel();

		// prsInst parses an instruction according to:
		//
		// <inst> ::= <opmove> <regofsval> "," <regofsval> |
		//						<opjmp> <regofs> |
		//						<opbitman> <regval> |
		//						<op2> <reg> "," <reg> |
		//						<op1> <reg> |
		//						<op0>
		void prsInst();

		// <reg> ::= "R00".."R31" plus special
		int prsReg();


		// *** INSTRUCTIONS ***
		
		// <op2reg> ::= <mnemonic_tworeg> <reg> "," <reg>
		void prs2reg();

		// <op1reg> ::= <mnemonic_onereg> <reg>
		void prs1reg();

		// <op1reg> ::= <mnemonic_zeroreg>
		void prs0reg();

		// <opbitman> ::= <mnemonic_bitman> <reg> "," <reg> | <mnemonic_bitman> <val> "," <reg>
		void prsBitman();

		// <opjmp> ::= <mnemonic_jump> <reg> | <mnemonic_jump> <id>
		void prsJmp();

		// <opmove> ::= "move" ...
		void prsMove();

		// parse move operand (difficult)
		moveOp prsMoveOperand( bool=true);



};

#endif