/////////////////////////////////////////////////////////////////////////////
//
// File: cpu.cpp
//
//
// The R1-emulator
// ---------------
//
// This is the master-file, defining all the code needed to emulate the
// R1-cpu. Make an instance of the class TCPU to execute R1-code.
//
// When ready, the load_obj() file may be used to load (and link) segments
// into memory from an object-file.
//
// Additional features may be added to the R1-emulator at a later date.
//

#include <string.h>
#include <iostream>
#include <iomanip>
#include <strstream>
#include <fstream>

#include "cpu.h"
#include "error.h"
#include "asm.h"

class ECPU: public Error {
public:
	enum ECode {
	 E_SUCCESS,
	 E_UNKNOWN,
	 E_ACCESSVIOLATION,
	 E_INVALIDINST,
	 E_NOTSUPPORTED
	};

	ECPU( ECode ec, string s ) {
		ecode = ec;
		emsg = s;
	}

	ECPU( ECode ec, WORD addr ) {
		ostrstream o;
		o << hex << setfill('0') << "0x" << addr << ends;
		emsg = o.str();
	}

	string getMsg() {
		ostrstream m;
		switch (ecode) {
			case E_SUCCESS: m << "no error" << ends; break;
			case E_UNKNOWN: m << "?" << ends; break;
			case E_ACCESSVIOLATION: m << "access violation at " << emsg << ends; break;
			case E_INVALIDINST: m << "invalid instruction: " << emsg << ends; break;
			case E_NOTSUPPORTED: m << "encountered unsupported instruction at: " << emsg << ends; break;
			default: m << "cpu error at " << emsg << ends;
		}
		return m.str();
	}

};


// Start default CPU emulator, registers cleared, all timers created and defaulted to 0
CPU::CPU( WORD memsize ): iregs((INT*)regs), state(RUNNING), cycleCount(0) {
	mem_size = memsize;
	memory = new BYTE[mem_size];
	memset(memory, 0, mem_size*sizeof(BYTE));
	for (int lp1=0; lp1<32; lp1++)
		regs[lp1] = 0;
	irq = new IRQDecoder();
	timer0 = new Timer(irq, INT_TIMER0);
	timer1 = new Timer(irq, INT_TIMER1);
	timer2 = new Timer(irq, INT_TIMER2);
	timer3 = new Timer(irq, INT_TIMER3);
}


// fname expected to be an object file in runes.object.format
// It will be loaded and relocated, so that bp points to data-
// segment, ip points to first byte in code, and r08-r16 points
// to the segments in the same order as they are loaded.
CPU::CPU( string fname ): iregs((INT*)regs), state(RUNNING), cycleCount(0) {
	fstream obj(fname.c_str(), ios::in | ios::binary );
	cout << "Loading from file: " << fname << endl;
	imageFile = fname;

	// Initialise all registers to 0
	for (int lp1=0; lp1<32; lp1++)
		regs[lp1] = 0;

	// allocate enough memory for the entire file
	obj.seekg(0, ios::end);
	mem_size = obj.tellg() / 2;
	obj.seekg(0);
	memory = new BYTE[mem_size];
	memset(memory, 0, mem_size*sizeof(BYTE));

	WORD mempos = 0;
	int rp = 8;

	while (!obj.eof()) {
		SegHeader segh;
		memset(&segh,0,sizeof(segh));
		obj.read(&segh, sizeof(segh));
		string segname("");

		for (int lp1=0; lp1<32 && segh.name[lp1]; lp1++)
			segname += (char) segh.name[lp1];
		// if segment is code, set R_IP to point to code segment
		if (segname=="code") R_IP=mempos; else
		// if segment is data, set R_BP to point to data segment
		if (segname=="data") R_BP=mempos; else
		// if segment is stack, set R_SP to point to top of stack
		if (segname=="stack") R_SP=mempos+segh.length; else
		// just set R8-R31 to point to segment 
			regs[rp++] = mempos;

		obj.read(memory+mempos, segh.length*sizeof(BYTE) );
		mempos += segh.length;
		cout << "Read segment: " << segname << endl;
	}
	obj.close();

	irq = new IRQDecoder();
	timer0 = new Timer(irq, INT_TIMER0);
	timer1 = new Timer(irq, INT_TIMER1);
	timer2 = new Timer(irq, INT_TIMER2);
	timer3 = new Timer(irq, INT_TIMER3);
}


CPU::~CPU() {
	if (memory) delete[] memory;
	if (irq) delete irq;
	if (timer0) delete timer0;
	if (timer1) delete timer1;
	if (timer2) delete timer0;
	if (timer3) delete timer1;	
}



#define SCREEN_BUFFER_LOC 0x0ffffe800L


// Read a word from memory
inline WORD CPU::memw( WORD address ) {
	BYTE *p = memory+address;
	if (address>=mem_size)
		throw ECPU(ECPU::E_ACCESSVIOLATION, address);
	return ((WORD)*(p+0)<<16) | ((WORD)*(p+1)<<0);
}

// Read a half-word from memory
inline BYTE CPU::memb( WORD address ) {
	BYTE *p = memory+address;
	if (address>mem_size)
		throw ECPU(ECPU::E_ACCESSVIOLATION, address);
	return (BYTE)*(p+0);
}

// Write a word to memory
inline void CPU::wmemw( WORD address, WORD value ) {
	if (address>=mem_size)
		throw ECPU(ECPU::E_ACCESSVIOLATION, address);
	*( (memory+address) ) = value >> 16;
	*( (memory+address+1) ) = value & 0x0FFFF;
}

// Write a half-word to memory
inline void CPU::wmemb( WORD address, BYTE value ) {
	if (address>mem_size)
		throw ECPU(ECPU::E_ACCESSVIOLATION, address);
	*((memory+address)) = value & 0x0FFFF;
}



WORD inline CPU::alu_add( WORD a, WORD b, bool sub, bool cin ) {
#ifndef INTEL_386
	// two's complement when subtracting
	if (sub) a = ~a;
	// carry is inverted if sub to become borrow
	cin = cin ^ sub;
	
	// calculate propagate/generate for all
	WORD propagate = a | b;
	WORD generate = a & b;
	WORD res = a + b + (cin?1:0);

	// correct for carry-in (generate[0] = (a[0] & b[0]) | ( cin & (a[0] | b[0]) )
	if (cin & (propagate & 1)) generate |= 1;

	// ASSUME LEFT-SHIFT INSERTS 0!
	bool msbprop = (propagate & 0x80000000); propagate <<= 1;
	bool msbgen = (generate & 0x80000000); generate <<= 1;
	bool msbcarry;
	while ( (propagate & 0x80000000) && !(generate & 0x80000000) ) {
		propagate <<= 1;
		generate <<= 1;
	}
	msbcarry = generate & 0x80000000;
	R_FLAGS.carry = msbgen || (msbprop && msbcarry);
	R_FLAGS.overflow = R_FLAGS.carry ^ msbcarry;
	R_FLAGS.sign = (res & 0x80000000) > 0;
	R_FLAGS.zero = res == 0;
	return res;
#endif
// currently the INTEL_386 does not work correctly
#ifdef INTEL_386
	WORD res;
	bool ovf=false, sign=false, zero=false, carry=false;
	asm ("
		mov eax, a
		mov ebx, b
		mov ch, sub
		mov cl, cin
		cmp ch, 0
		je	add
		xor cl, 0xff
		not ebx
	sub:
		btt cl, 0
		sbb eax, ebx
		jmp updateflags
	add:
		btt cl, 0
		adc eax, ebx
	updateflags:
		mov res, eax
		jnz zzz
		mov zero, 1
		zzz:
		jno ooo
		mov ovf, 1
		ooo:
		jns sss
		mov sign, 1
		sss:
		jc	ccc
		mov carry, 1
		ccc:
	");
	R_FLAGS.carry = carry;
	R_FLAGS.overflow = ovf;
	R_FLAGS.sign = sign;
	R_FLAGS.zero = zero;
	return res;

#endif
}


// Execute one instruction
// Return value is size of instruction
int CPU::step() {
	Opcode inst;
	WORD temp;
	WORD oldip = R_IP;
	int icycles = 1; // currently, all instructions are 1 cycle long
	
	// look for interrupts before execing instructions
	int i = irq->find();
	if (i>=0) {
		// an interrupt was found and may be serviced
		if ((R_FLAGS.intrenable && i>=8) || (i<8)) {
			// transfer execution since not masked or unmaskable
			state = RUNNING;
			irq->start(i);
			R_SP -= 2;
			wmemw( R_SP, R_FL );
			R_SP -= 2;
			wmemw( R_SP, R_IP );			
			R_FLAGS.intrenable = 0;
			R_FLAGS.singlestep = 0;
			R_IP = memw( R_IT+2*i );
		}
	}			

	if (state == HALTED) {
		cycleCount++;
		return 0;
	}
	// Read instruction from memory
	inst.opcode = memb(R_IP);
	R_IP += 1;


	// Trap 0-reg instructions
	if ((inst.opcode&0x0ffe0)==0x0ffe0) {

		switch ( inst.zeroReg.inst ) {
			case OC_HALT:
				R_FLAGS.intrenable = 1;
				state = HALTED;
				break;

			case OC_RET:
				R_IP = memw(R_SP);
				R_SP += 2;
				break;

			case OC_IRET:
				R_IP = memw(R_SP);
				R_SP += 2;
				R_FL = memw(R_SP);
				R_SP += 2;
				irq->done();
				break;

			case OC_CALL:
				temp = memw(R_IP);
				R_IP += 2;
				R_SP -= 2;
				wmemw(R_SP, R_IP);
				R_IP += temp;
				break;

			case OC_JMP:
				temp = memw(R_IP);
				R_IP += 2;
				R_IP += temp;
				break;

			case OC_JE:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.zero) R_IP += temp;
				break;

			case OC_JNE:
				temp = memw(R_IP);
				R_IP += 2;
				if (!R_FLAGS.zero) R_IP += temp;
				break;

			case OC_JC:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.carry) R_IP += temp;
				break;

			case OC_JNC:
				temp = memw(R_IP);
				R_IP += 2;
				if (!R_FLAGS.carry) R_IP += temp;
				break;

			case OC_JS:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.sign) R_IP += temp;
				break;

			case OC_JNS:
				temp = memw(R_IP);
				R_IP += 2;
				if (!R_FLAGS.sign) R_IP += temp;
				break;

			case OC_JA:
				temp = memw(R_IP);
				R_IP += 2;
				if (!(R_FLAGS.carry || R_FLAGS.zero)) R_IP += temp;
				break;

			case OC_JAE:
				temp = memw(R_IP);
				R_IP += 2;
				if (!R_FLAGS.carry) R_IP += temp;
				break;

			case OC_JB:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.carry) R_IP += temp;
				break;

			case OC_JBE:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.carry || R_FLAGS.zero) R_IP += temp;
				break;

			case OC_JG:
				temp = memw(R_IP);
				R_IP += 2;
				if (!((R_FLAGS.sign^R_FLAGS.overflow) || R_FLAGS.zero)) R_IP += temp;
				break;

			case OC_JGE:
				temp = memw(R_IP);
				R_IP += 2;
				if (!(R_FLAGS.sign^R_FLAGS.overflow)) R_IP += temp;
				break;

			case OC_JL:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.sign^R_FLAGS.overflow) R_IP += temp;
				break;

			case OC_JLE:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.zero||(R_FLAGS.sign^R_FLAGS.overflow)) R_IP += temp;
				break;

			case OC_JO:
				temp = memw(R_IP);
				R_IP += 2;
				if (R_FLAGS.overflow) R_IP += temp;
				break;

			default: 
				irq->request(INT_OPCODE);
		}

	} else
	// Trap 1-reg instructions
	if ((inst.opcode&0x0fc00)==0x0fc00) {
		switch ( inst.oneReg.inst ) {

			case OC_CLR:
				regs[ inst.oneReg.reg ] = 0;
				R_FLAGS.zero = 1;
				R_FLAGS.sign = 0;
				break;

			case OC_SET:
				regs[ inst.oneReg.reg ] = 0x0ffffffffL;
				R_FLAGS.zero = 0;
				R_FLAGS.sign = 1;
				break;

			case OC_JMP:
				R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_CALL:
				R_SP -= 2;
				wmemw( R_SP, R_IP );
				R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_PUSH:
				R_SP -= 2;
				wmemw( R_SP, regs[ inst.oneReg.reg ] );
				break;

			case OC_POP:
				regs[ inst.oneReg.reg] = memw( R_SP );
				R_SP += 2;
				R_FLAGS.zero = regs[ inst.oneReg.reg ]==0;
				R_FLAGS.sign = regs[ inst.oneReg.reg ]>>31;
				break;

			case OC_INC:
				//regs[ inst.oneReg.reg ]++;
				regs[ inst.oneReg.reg ] = alu_add(0, regs[ inst.oneReg.reg ], false, true);
				break;

			case OC_DEC:
				//regs[ inst.oneReg.reg ]--;
				regs[ inst.oneReg.reg ] = alu_add(0, regs[ inst.oneReg.reg ], true, true);
				break;

			case OC_NOT:
				regs[ inst.oneReg.reg] = ~regs[ inst.oneReg.reg];
				R_FLAGS.zero = regs[ inst.oneReg.reg ]==0;
				R_FLAGS.sign = regs[ inst.oneReg.reg ]>>31;
				break;

			case OC_NEG:
				regs[ inst.oneReg.reg] = -regs[ inst.oneReg.reg];
				R_FLAGS.zero = regs[ inst.oneReg.reg ]==0;
				R_FLAGS.sign = regs[ inst.oneReg.reg ]>>31;
				break;

			case OC_LOADIMM:
				regs[ inst.oneReg.reg] = memw(R_IP+0);
				R_IP += 2;
				R_FLAGS.zero = regs[ inst.oneReg.reg ]==0;
				R_FLAGS.sign = regs[ inst.oneReg.reg ]>>31;
				break;

			case OC_LOADIMM_H:
				regs[ inst.oneReg.reg ] &= 0x0ffff0000L;
				regs[ inst.oneReg.reg ] |= memb(R_IP+0);
				R_IP += 1;
				R_FLAGS.zero = regs[ inst.oneReg.reg ]==0;
				R_FLAGS.sign = regs[ inst.oneReg.reg ]>>31;
				break;

			case OC_JE:
				if (R_FLAGS.zero) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JNE:
				if (!R_FLAGS.zero) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JC:
				if (R_FLAGS.carry) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JNC:
				if (!R_FLAGS.carry) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JS:
				if (R_FLAGS.sign) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JNS:
				if (!R_FLAGS.sign) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JA:
				if (!(R_FLAGS.carry || R_FLAGS.zero)) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JAE:
				if (!R_FLAGS.carry) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JB:
				if (R_FLAGS.carry) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JBE:
				if (R_FLAGS.carry || R_FLAGS.zero) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JG:
				if (!((R_FLAGS.sign^R_FLAGS.overflow) || R_FLAGS.zero)) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JGE:
				if (!(R_FLAGS.sign^R_FLAGS.overflow)) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JL:
				if (R_FLAGS.sign^R_FLAGS.overflow) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JLE:
				if (R_FLAGS.zero||(R_FLAGS.sign^R_FLAGS.overflow)) R_IP = regs[ inst.oneReg.reg ];
				break;

			case OC_JO:
				if (R_FLAGS.overflow) R_IP = regs[ inst.oneReg.reg ];
				break;


			default: 
				irq->request(INT_OPCODE);

		}

	} else
	// 2-reg instruction
	switch ( inst.twoReg.inst ) {

		case OC_NOP:
			break;

		case OC_MOVE:
			if (inst.twoReg.param1)
				temp = memw(regs[ inst.twoReg.source ] + R_BP);
			else
				temp = regs[ inst.twoReg.source ];
			if (inst.twoReg.param2)
				wmemw(regs[ inst.twoReg.dest ] + R_BP, temp);
			else
				regs[ inst.twoReg.dest ] = temp;
			R_FLAGS.zero = temp==0;
			R_FLAGS.sign = temp>>31;
			break;

		case OC_MOVE_H:
			if (inst.twoReg.param1)
				temp = memb(regs[ inst.twoReg.source ] + R_BP);
			else
				temp = regs[ inst.twoReg.source ];
			if (inst.twoReg.param2)
				wmemb(regs[ inst.twoReg.dest ] + R_BP, temp);
			else
				regs[ inst.twoReg.dest ] = temp;
			R_FLAGS.zero = temp == 0;
			R_FLAGS.sign = temp >> 15;
			break;


		case OC_MOVEBO: // param2=store (load=0) - param1=bytemove
			temp = memw( R_IP );
			R_IP += 2;
			if (inst.twoReg.param1) {
				// byte move:
				if (inst.twoReg.param2)	wmemb( regs[inst.twoReg.dest] + temp, regs[inst.twoReg.source] );
				else {
					regs[ inst.twoReg.dest ] = memb( regs[inst.twoReg.source] + temp );
					temp = regs[inst.twoReg.dest];
					R_FLAGS.zero = temp == 0;
					R_FLAGS.sign = temp >> 15;
				}
			} else {
				// full-word
				if (inst.twoReg.param2)	wmemw( regs[ inst.twoReg.dest ] + temp, regs[ inst.twoReg.source ]);
				else {
					regs[ inst.twoReg.dest ] = memw( regs[inst.twoReg.source ] + temp );
					temp = regs[inst.twoReg.dest];
					R_FLAGS.zero = temp == 0;
					R_FLAGS.sign = temp >> 31;
				}
			}
			break;



		case OC_ADD:
			if (inst.twoReg.param1)
				;
			else
				regs[ inst.twoReg.dest] = alu_add( regs[ inst.twoReg.source ], regs[ inst.twoReg.dest ], false, inst.twoReg.param2 && R_FLAGS.carry);
			break;


		case OC_SUB: // param1=store
			if (inst.twoReg.param1)
				alu_add( regs[ inst.twoReg.dest ], regs[ inst.twoReg.source ], true, inst.twoReg.param2 && R_FLAGS.carry);
			else
 				regs[ inst.twoReg.dest ] = alu_add( regs[ inst.twoReg.source ], regs[ inst.twoReg.dest ], true, inst.twoReg.param2 && R_FLAGS.carry);
			break;


		case OC_LOGICAL:
			temp = 0;
			if (inst.twoReg.param1)
				if (inst.twoReg.param2)
					temp = regs[inst.twoReg.dest] ^ regs[inst.twoReg.source];
				else
					temp = regs[inst.twoReg.dest] | regs[inst.twoReg.source];
			else
				if (inst.twoReg.param2)
					;
				else
					temp = regs[inst.twoReg.dest] & regs[inst.twoReg.source];
			regs[inst.twoReg.dest] = temp;
			R_FLAGS.zero = temp==0;
			R_FLAGS.sign = temp>>31;
			break;


		case OC_BITTEST:
			if ( inst.twoReg.param2 )
				R_FLAGS.carry = (regs[inst.twoReg.dest] >> inst.twoReg.source) & 1;
			else
				R_FLAGS.carry = (regs[inst.twoReg.dest] >> regs[inst.twoReg.source&31]) & 1;
			break;


		case OC_BITMAN:
			temp = 1;
			if ( inst.twoReg.param1 )
				temp = temp << inst.twoReg.source;
			else
				temp = temp << (regs[inst.twoReg.source]&31);
			if ( inst.twoReg.param2 )
				regs[ inst.twoReg.dest ] |= temp;
			else
				regs[ inst.twoReg.dest ] &= ~temp;

			R_FLAGS.zero = regs[ inst.twoReg.dest ]==0;
			R_FLAGS.sign = regs[ inst.twoReg.dest ]>>31;
			break;


		case OC_SHR:
			if ( !inst.twoReg.param1) {
				if ( inst.twoReg.param2 )
					regs[ inst.twoReg.dest ] >>= inst.twoReg.source;
				else
					regs[ inst.twoReg.dest ] >>= regs[inst.twoReg.source&31];
			} else {
				if ( inst.twoReg.param2 )
					(INT) regs[ inst.twoReg.dest ] >>= inst.twoReg.source;
				else
					(INT) regs[ inst.twoReg.dest ] >>= regs[inst.twoReg.source&31];
			}
			break;


		case OC_SHL:
			if ( inst.twoReg.param2 )
				regs[ inst.twoReg.dest ] <<= inst.twoReg.source;
			else
				regs[ inst.twoReg.dest ] <<= regs[inst.twoReg.source&31];
			break;


		default: 
			irq->request(INT_OPCODE);

		case OC_ROR:
			break;


		case OC_ROL:
			break;

	}

	// update global cycle-counter
	cycleCount += icycles;
	// update internal timers
	for (;icycles; icycles--) {
		if (timer0) timer0->clk();
		if (timer1) timer1->clk();
		if (timer2) timer2->clk();
		if (timer3) timer3->clk();
	}
	// if singlestep, request BRKPNT interrupt
	if (R_FLAGS.singlestep) 
		irq->request(INT_BRKPNT);
	
	return R_IP-oldip;
}


void CPU::run() {
	while (state==RUNNING) step();
}