/////////////////////////////////////////////////////////////////////////////
//
// File: disasm.cpp
//
// Disassembler for R1-emulator
// ----------------------------
//
// Pass pointer to disasm(), the instruction is disassembled and
// written to the ostream, you specify.
// Pass pointer and count to disAsmMem() and memory-area is disassembled
// and result is written to the ostream, you specify.
//
#include "disasm.h"
#include <iomanip>
#include <strstream>
#include <string>

// this table contains the strings for all the registers
static char* str_reg[] =
  { "ip",  "fl",  "sp",  "bp",  "it",  "r--", "ptb", "psz",
    "r08", "r09", "r10", "r11", "r12", "r13", "r14", "r15",
    "r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
    "r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31"  };


// The string supplied in out is filled with disassembled code.
// The return-value is the length of the instruction.
// NOTE: If addr!=0 then indexing into memory!
int DisAsm::disasm( ostream& out, BYTE *mem, WORD addr=0 ) {
  // the 3 added to a+i is to correct for the movement of IP due to immediate values
  #define WRITE_IMM_OFF(a,i) (a!=0 ? out << hex << (a+i+3) << "h" << ends : out << str_reg[0] << "+" << dec << (INT)imm << ends)

	union Opcode inst;
  char str[40];

  mem += addr;
	inst.opcode = *mem;

	/* Trap 0-reg instructions */
	if (inst.zeroReg.reserved==0x07ff) {
		switch ( inst.zeroReg.inst ) {

      case OC_HALT:
				out << "halt" << ends;
				return 1;

			case OC_RET:
				out << "ret" << ends;
				return 1;

			case OC_IRET:
				out << "iret" << ends;
				return 1;

			case OC_JMP: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jmp  ";
        WRITE_IMM_OFF(addr,imm);
        return 3;
			}

			case OC_CALL: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "call ";
        WRITE_IMM_OFF(addr,imm);
        return 3;
			}

			case OC_JE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "je   ";
        WRITE_IMM_OFF(addr,imm);
        return 3;
			}

			case OC_JNE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jne  ";
        WRITE_IMM_OFF(addr,imm);
        return 3;
			}

			case OC_JC: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jc   ";
        WRITE_IMM_OFF(addr,imm);
        return 3;
			}

			case OC_JNC: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jnc  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JS: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "js   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JNS: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jns  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JA: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "ja   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JAE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jae  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JB: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jb   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JBE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jbe  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JG: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jg   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JGE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jge  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JL: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jl   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JLE: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jle  ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}

			case OC_JO: {
				WORD imm = *(mem+1)<<16 | *(mem+2);
        out << "jo   ";
        WRITE_IMM_OFF(addr,imm);
				return 3;
			}


			default:
				out << "???" << ends;
				return 1;

		}

	} else
	/* Trap 1-reg instructions */
	if (inst.oneReg.reserved == 0x03f) {
		switch ( inst.oneReg.inst ) {

      case OC_CLR:
        out << "clr  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_SET:
        out << "set  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_PUSH:
        out << "push " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_POP:
        out << "pop  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_INC:
        out << "inc  " << str_reg[inst.oneReg.reg] << ends;
        out << str << ends;
				return 1;

			case OC_DEC:
        out << "dec  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_NOT:
        out << "not  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_NEG:
        out << "neg  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_LOADIMM: {
				WORD imm = ((WORD)*(mem+1)<<16) | (WORD)*(mem+2);
        out << "move.i " << dec << imm << ", " << str_reg[inst.oneReg.reg] << ends;
				return 3;
			}

			case OC_LOADIMM_H: {
				BYTE imm = *(mem+1);
        out << "move.ib " << dec << imm << ", " << str_reg[inst.oneReg.reg] << ends;
				return 2;
      }

			case OC_JMP:
        out << "jmp  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_CALL:
        out << "call " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JE:
        out << "je   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JNE:
        out << "jne  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JC:
        out << "jc   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JNC:
        out << "jnc  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JS:
        out << "js   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JNS:
        out << "jns  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JA:
        out << "ja   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JAE:
        out << "jae  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JB:
        out << "jb   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JBE:
        out << "jbe  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JG:
        out << "jg   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JGE:
        out << "jge  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JL:
        out << "jl   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JLE:
        out << "jle  " << str_reg[inst.oneReg.reg] << ends;
				return 1;

			case OC_JO:
        out << "jo   " << str_reg[inst.oneReg.reg] << ends;
				return 1;

      default:
        out << "???  " << str_reg[inst.oneReg.reg] << ends;
        return 1;

		}

	} else
	/* 2-reg instruction */
	{
		switch ( inst.twoReg.inst ) {

      case OC_NOP:
        out << "nop" << ends;
				return 1;

			case OC_MOVE:
        out << "move ";
				if (inst.twoReg.param1) out << "bp:[" << str_reg[inst.twoReg.source] << "], ";
        else out << str_reg[inst.twoReg.source] << ", ";
				if (inst.twoReg.param2) out << "bp:[" << str_reg[inst.twoReg.dest] << "]";
        else out << str_reg[inst.twoReg.dest];
        out << ends;
				return 1;

			case OC_MOVE_H:
        out << "move.b ";
				if (inst.twoReg.param1) out << "bp:[" << str_reg[inst.twoReg.source] << "], ";
        else out << str_reg[inst.twoReg.source] << ", ";
				if (inst.twoReg.param2) out << "bp:[" << str_reg[inst.twoReg.dest] << "]";
        else out << str_reg[inst.twoReg.dest];
        out << ends;
				return 1;


			case OC_MOVEBO: {
				WORD imm = (WORD)*(mem+1) << 16 | *(mem+2);
        out << (inst.twoReg.param1?"move.b ":"move ");

				if (inst.twoReg.param2)
          out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ":[" << hex << (INT)imm << "h]";
        else
          out << str_reg[inst.twoReg.source] << ":[" << hex << (INT)imm << "h], " << str_reg[inst.twoReg.dest];
        out << ends;
				return 3;
			}

			case OC_ADD:
				if ( inst.twoReg.param1 &&  inst.twoReg.param2) out << "???add ";
				if ( inst.twoReg.param1 && !inst.twoReg.param2) out << "???add ";
				if (!inst.twoReg.param1 &&  inst.twoReg.param2) out << "adc  ";
				if (!inst.twoReg.param1 && !inst.twoReg.param2) out << "add  ";
        out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_SUB:
				if ( inst.twoReg.param1 &&  inst.twoReg.param2) out << "???cmp ";
				if ( inst.twoReg.param1 && !inst.twoReg.param2) out << "cmp  ";
				if (!inst.twoReg.param1 &&  inst.twoReg.param2) out << "sbb  ";
				if (!inst.twoReg.param1 && !inst.twoReg.param2) out << "sub  ";
        out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_LOGICAL:
				if ( inst.twoReg.param1 &&  inst.twoReg.param2) out << "xor  ";
				if ( inst.twoReg.param1 && !inst.twoReg.param2) out << "or   ";
				if (!inst.twoReg.param1 &&  inst.twoReg.param2) out << "test ";
				if (!inst.twoReg.param1 && !inst.twoReg.param2) out << "and  ";
        out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_BITTEST:
				out << (inst.twoReg.param1? "btr  " : "btt  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_BITMAN:
				out << (inst.twoReg.param1? "bts  " : "btc  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_SHR:
				out << (inst.twoReg.param1? "sar  " : "shr  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_SHL:
				out << (inst.twoReg.param1? "???sal  " : "shl  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_ROR:
				out << (inst.twoReg.param1? "rcr  " : "ror  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

			case OC_ROL:
				out << (inst.twoReg.param1? "rcl  " : "rol  ");
        if (inst.twoReg.param2) out << inst.twoReg.source << ", " << str_reg[inst.twoReg.dest] << ends;
        else out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

      case OC_MULDIV:
        out << "???";
				out << (inst.twoReg.param1? "i" : "");
        out << (inst.twoReg.param2? "div ":"mul ");
				out << (inst.twoReg.param1? "" : " ");
        out << str_reg[inst.twoReg.source] << ", " << str_reg[inst.twoReg.dest] << ends;
				return 1;

		}

	}

  out << hex << "??? - [15-10]:" << (inst.opcode>>10) <<" [9-5]:" << ((inst.opcode>>5)&0x1f) << " [4-0]:" << (inst.opcode&0x1f) << ends;
	return 1;
}



// A function for doing left-aligns
static string lalign(int isize, string i, char fill=' ') {
  string res;
  int lp1;
  for (lp1=0; lp1<isize; lp1++)
     res += (lp1<i.length()?i[lp1]:fill);
  return res;
}


// An entire memory-area is disassembled and the output is written to
// the stream supplied in out
void DisAsm::disAsmMem( BYTE *mem, WORD size, ostream& out ) {
	int instsize = 1;
	WORD pos = 0;

	while ((signed)size>0 && instsize!=0) {
    ostrstream da;
    ostrstream opcodes;
		instsize = disasm(da, mem, pos);
    out << hex << setw(8) << setfill('0') << pos << "h:  ";

    for (int lp1=0; lp1<instsize; lp1++)
      opcodes << hex << setw(4) << setfill('0') << *(mem+pos+lp1) << " ";
    opcodes << ends;

    out << lalign(14,opcodes.str()) << "  " << da.str() << endl;
		pos += instsize;
		size -= instsize;
	}
}