///////////////////////////////////////////////////////////////////////////////
// File: emulator.cpp
// ------------------
// master file for the emulator

#include "cpu.h"
#include "error.h"
#include "console.h"
#include <stdio.h>
#include <strstream>
#include <iomanip>
#include "disasm.h"
#include "scanner.h"
#include "dos.h"

class EmuCPU: public CPU {
	public:
	EmuCPU(string fname);
	~EmuCPU();
	
	int run();
	void ui();
	void redraw_screen();
	void write_emustate();
	
	Console *con;
	ConEdit *prompt;
	DisAsm dasm;
	
};
	
EmuCPU::EmuCPU(string fname): CPU(fname), con(NULL), prompt(NULL) {
	con = new Console();
	prompt = new ConEdit(con,4,22,false,72,0x70);
}

EmuCPU::~EmuCPU() {
	if (prompt) delete prompt;
	if (con) delete con;
}
	 
	
// Draw all fixed-info fields on screen - this includes headers, footers, register-names, etc.
// registers are located at: 3-19 (16 lines, two on each line
void EmuCPU::redraw_screen() {
	char* str_reg[] =	{ 
		"ip ", "fl ", "sp ", "bp ", "it ", "r--", "ptb", "psz",
		"r08", "r09", "r10", "r11", "r12", "r13", "r14", "r15",
		"r16", "r17", "r18", "r19", "r20", "r21", "r22", "r23",
		"r24", "r25", "r26", "r27", "r28", "r29", "r30", "r31" };
		
	con->setRowCount(50);
	con->textattr(0x17); 
	con->clrscr();

	// write header
	con->gotoxy(1,1);
	con->textattr(0x70); con->clreol();
	con->print("  -  R1-emulator (w) 1999-2000 Rune Barnkob @ MeeSoft : ");
	con->printstring( imageFile );
	// write footer
	con->gotoxy(1,50);
	con->textattr(0x70); con->clreol();
	con->print(" emustate of cpu on this screen. Enter command");

	con->textattr(0x1F);
	
	// write fixed-register names
	for (int lp1=0, lp2=16; lp1<16; lp1++, lp2++) {
		con->gotoxy(1, 3+lp1);
		con->print(str_reg[lp1]);
		con->gotoxy(40, 3+lp1);
		con->print(str_reg[lp2]);
	}
	
	// write prompt
	con->gotoxy(1,22);
	con->print("==> ");
}


// Shows the CPU-emustate
void EmuCPU::write_emustate() {
	char reg_str[80];

	con->textattr(0x17);
	
	// update fields with register contents
	for (int lp1=0, lp2=16; lp1<16; lp1++, lp2++) {
		sprintf(reg_str, "%08lxh = %10ld", regs[lp1], regs[lp1]);
		con->gotoxy(6, 3+lp1); con->print(reg_str);
		sprintf(reg_str, "%08lxh = %10ld", regs[lp2], regs[lp2]);
		con->gotoxy(46, 3+lp1); con->print(reg_str);
	}

	// write flags
	sprintf(reg_str, "i=%d    o=%d    s=%d    c=%d    z=%d    step=%d",
			 R_FLAGS.intrenable,
			 R_FLAGS.overflow,
			 R_FLAGS.sign,
			 R_FLAGS.carry,
			 R_FLAGS.zero,
			 R_FLAGS.singlestep);
	con->gotoxy(1, 19); 
	con->print(reg_str);

	// 25->45 reserved for showing diff. info - default is disassembled code
	for (int lp1=0; lp1<20; lp1++) {
		 con->gotoxy(1,25+lp1); 
		 con->clreol();
	}

	WORD pos = R_IP;
	for (int lp1=0; lp1<20; lp1++) {
		ostrstream da;
		ostrstream opcd;

		int insz = dasm.disasm(da, memory, pos );
		opcd << hex << setw(8) << setfill('0') << pos << "h: ";
		for (int lp2=0; lp2<insz; lp2++)
			opcd << hex << setw(4) << setfill('0') << *(memory+pos+lp2) << " ";
		opcd << ends;
		con->gotoxy(1,25+lp1);
		con->printstring( opcd.str() );
		con->gotoxy(40,25+lp1);
		con->printstring( da.str() );
		pos += insz;
	}
	
	// print IRQ state
	{
		ostrstream o;
		irq->state(o);
		o << ends;
		con->gotoxy(1,21);
		con->print(o.str());
	}
	
	
	// print current cycle count
	sprintf(reg_str, "%9d", cycleCount);
	con->gotoxy(65,2);
	con->print(reg_str);
	con->textattr(0x4F);
	// print current CPU state
	con->gotoxy(75,2);
	switch (state) {
		case CPU::HALTED: con->print("HALTED"); break;
		case CPU::RUNNING: con->print("RUN   "); break;
	}
	return;
}


// Run program in CPU
// Starts executing from IP
int EmuCPU::run() {
	redraw_screen();
	write_emustate();
	try {
		ui();
	} catch (Error& e) {
		cerr << "!!! Error: " << e.getMsg() << endl;
		ui();
	}
	cout << "...emulation terminated..." << endl;
}

// User interface - show emustate - wait for command
void EmuCPU::ui() {
	enum op { EXIT, STEP, RUN, ANI, TBL, INT } o;
	char buf[80];
	Scanner sc(false);
	
	
	// look to see if user has pressed key or we are in stepping-mode
	while (true) {
		write_emustate();
		buf[0] = 0;
		prompt->setText(buf);
		// read command
		while (prompt->process()!=kpDone);
		prompt->getText(buf);
		sc.newLine(buf);
		sc.next();
		if (sc.curTok()==TT_ID) {
			if (sc.getLex()=="exit" || sc.getLex()=="x") o = EXIT;
			if (sc.getLex()=="run" || sc.getLex()=="r") o = RUN;
			if (sc.getLex()=="step" || sc.getLex()=="s") o = STEP;
			if (sc.getLex()=="ani" || sc.getLex()=="a") o = ANI;
			if (sc.getLex()=="mem") o = TBL;
			if (sc.getLex()=="int") o = INT;
		} else o = STEP;
		
		switch (o) {
			case EXIT:
				return;
				
			case INT: {
				WORD irqno;
				con->textattr(0x70); 
				con->gotoxy(1,50); 
				con->clreol(); 
				con->print("requestingg interrupts...");
				sc.next();
				while (sc.curTok()==TT_CONST) {
					irqno=sc.getWord();
					irq->request(irqno);
					sc.next();
				}
				break;
			}
				
				
			case RUN: 
				con->textattr(0x70); 
				con->gotoxy(1,50); 
				con->clreol(); 
				con->print("running...press any key to break");
				while (!con->keyhit() && state!=CPU::HALTED && !R_FLAGS.singlestep) 
					step();
				break;
				
			case STEP: 
				step();
				con->textattr(0x70); 
				con->gotoxy(1,50); 
				con->clreol(); 
				con->print("one step ahead");
				break;
				
			case ANI: {
				WORD msec = 200;
				sc.next();
				if (sc.curTok()==TT_CONST) msec=sc.getWord();
				con->textattr(0x70); 
				con->gotoxy(1,50); 
				con->clreol(); 
				con->print("animating...press any key to break");
				while (!con->keyhit() && state!=CPU::HALTED && !R_FLAGS.singlestep) {
					step();
					delay(msec);
					write_emustate();
				}
				break;
			}

			case TBL: {
				WORD pos = R_BP;
				WORD lines = 20;
				con->textattr(0x70); 
				con->gotoxy(1,50); 
				con->clreol(); 
				con->print("memory dump...");
				
				sc.next();
				if (sc.curTok()==TT_CONST) {
					pos = sc.getWord();
					sc.next();
					if (sc.curTok()==TT_CONST) {
						lines = sc.getWord();
					}
				}
				if (lines>20) lines=20;
				
				con->textattr(0x17);
				for (int lp1=0; lp1<lines; lp1++) {
					con->gotoxy(1,25+lp1);
					con->clreol();
				}
				
				for (int lp1=0; lp1<lines; lp1++) {
					ostrstream mem;
					
					con->gotoxy(1,25+lp1);
					mem << hex << setw(8) << setfill('0') << pos << "h:    ";
					for (int lp2=0; lp2<8; lp2++) {
						if (pos<mem_size)
							mem << hex << setw(4) << setfill('0') << *(memory+pos++) << " ";
						else
							mem << "---- ";
					}
					mem << ends;
					con->print(mem.str());
				}
				while (!con->keyhit()) ;
			}
		}

	}
}




#include <iostream>

int main( int argc, char *argv[] ) {
	if (argc<2) {
		cout << "R1-emulator" << endl;
		cout << "-----------" << endl;
		cout << "Must supply object name!" << endl;
		return 1;
	}
	string fname = argv[1];

	try {
		EmuCPU cpu(fname);
		cpu.run();
		return 0;
	} catch (Error& e) {
		cerr << e.getMsg() << endl;
		return 1;
	}
}
