/////////////////////////////////////////////////////////////////////////////
//
// File: scanner.cpp
//
//
// Scanner for the R1-assembler:
// -----------------------------
//	ident		::= <alpha> { <alphanum> }
//	number	 ::= "0" <base-select> { <hexdigit> } | <digit> { <digit> }
//
//	A series of keywords derived from the instruction mnemonics.

#include "scanner.h"
#include <ctype.h>
#include "error.h"
#include <strstream>
#include <iomanip>


class EScanner: public Error {
public:
	enum { E_SUCCESS=0, E_UNKNOWN, E_INVALIDCHAR, E_MISSINGQUOTE, E_UNKNOWNESC, E_DIGITLARGERBASE };

	EScanner(int se): Error(se) {}

	EScanner(int se, int chno ): Error(se), charno(chno) {}


	string getMsg() {
		ostrstream msg;
		msg << "SC" << setw(3) << setfill('0') << ecode << ": ";

		switch (ecode) {
			case E_SUCCESS: msg << "no error" << ends; break;
			case E_UNKNOWN: msg << "unknown error" << ends; break;
			case E_INVALIDCHAR: msg << "invalid character at " << charno << ends; break;
			case E_MISSINGQUOTE: msg << "missing \" at end of string" << ends; break;
			case E_UNKNOWNESC: msg << "unknown escape-sequence at " << charno << ends; break;
			case E_DIGITLARGERBASE: msg << "the digit at " << charno << " is larger than the base for the number" << ends; break;
		}
		return msg.str();
	}

protected:
	int charno;

};


#define DEBUG(s) //(cerr << s << endl)

// Global keywords map. Initialised by constructor
static map<string, int> keywords;


// Constructor for scanner object. Initialises all keywords in the table
Scanner::Scanner(bool checkkeywords = true): chk_kw(checkkeywords) {
	// instructions
	keywords["nop"] = TT_NOP;
	keywords["add"] = TT_ADD;
	keywords["adc"] = TT_ADC;
	keywords["sub"] = TT_SUB;
	keywords["sbb"] = TT_SBB;
	keywords["cmp"] = TT_CMP;
	keywords["and"] = TT_AND;
	keywords["test"]= TT_TEST;
	keywords["or"]	 = TT_OR;
	keywords["xor"] = TT_XOR;
	keywords["btt"] = TT_BTT;
	keywords["btr"] = TT_BTR;
	keywords["btc"] = TT_BTC;
	keywords["bts"] = TT_BTS;
	keywords["shr"] = TT_SHR;
	keywords["sar"] = TT_SAR;
	keywords["shl"] = TT_SHL;
	keywords["sal"] = TT_SHL;
	keywords["ror"] = TT_ROR;
	keywords["rcr"] = TT_RCR;
	keywords["rol"] = TT_ROL;
	keywords["rcl"] = TT_RCL;
	keywords["mul"] = TT_MUL;
	keywords["imul"]= TT_IMUL;
	keywords["div"] = TT_DIV;
	keywords["idiv"]= TT_IDIV;
	keywords["move"]= TT_MOVE;
	keywords["clr"] = TT_CLR;
	keywords["set"] = TT_SET;
	keywords["not"] = TT_NOT;
	keywords["neg"] = TT_NEG;
	keywords["inc"] = TT_INC;
	keywords["dec"] = TT_DEC;
	keywords["push"]= TT_PUSH;
	keywords["pop"] = TT_POP;
	keywords["jmp"] = TT_JMP;
	keywords["call"]= TT_CALL;
	keywords["je"]	 = TT_JE;
	keywords["jz"]	 = TT_JE;		// jz is the same as je
	keywords["jne"] = TT_JNE;
	keywords["jnz"] = TT_JNE;	// jnz is the same as jne
	keywords["jc"]	 = TT_JC;
	keywords["jnc"] = TT_JNC;
	keywords["js"]	 = TT_JS;
	keywords["jns"] = TT_JNS;
	keywords["ja"]	 = TT_JA;
	keywords["jae"] = TT_JAE;
	keywords["jb"]	 = TT_JB;
	keywords["jbe"] = TT_JBE;
	keywords["jg"]	 = TT_JG;
	keywords["jge"] = TT_JGE;
	keywords["jl"]	 = TT_JL;
	keywords["jle"] = TT_JLE;
	keywords["ret"] = TT_RET;
	keywords["iret"]= TT_IRET;
	keywords["halt"]= TT_HALT;
	// directives
	keywords["seg"] = TT_SEG;
	keywords["end"] = TT_END;
	keywords["assume"] = TT_ASSUME;
	// five regs
	keywords["ip"] = TT_REG;
	keywords["fl"] = TT_REG;
	keywords["flags"] = TT_REG;
	keywords["sp"] = TT_REG;
	keywords["bp"] = TT_REG;
	keywords["it"] = TT_REG;
	// storage specifiers
	keywords["dw"] = TT_DW;
	keywords["db"] = TT_DB;

	line[0] = 0;
	linepos = 0;
	tok = TT_EOL;
	lex = "";
}


// Destructor
Scanner::~Scanner() {}


// Initialise scanner for a new line
void Scanner::newLine( string newline ) {
	DEBUG("new line inited: " << newline );
	for (int lp1=0; lp1<256; lp1++)
		line[lp1] = ( lp1>=newline.length() ? 0 : tolower(newline[lp1]) );
	linepos = 0;
	lex = "";
	tok = TT_EOL;
}


// Return Lexeme
string Scanner::getLex() {
	return lex;
}

// Return Lexeme
WORD Scanner::getWord() {
	return val;
}

// get next token
ScTokens Scanner::next() {
	lex = "";
	tok = TT_EOL;


	// skip whitespaces
	while (isspace(line[linepos]))
		linepos++;
	
	// stop if end-of-line
	if (line[linepos]==0) return TT_EOL;

	// watch for line-comments
	if (line[linepos]=='#') return TT_EOL;


	// if digit then read number
	if (isdigit(line[linepos])) {
		scVal();
		return tok;
	}

	// if letter then read identifier
	if (isalpha(line[linepos]) || line[linepos]=='_') {
		scId();
		return tok;
	}
	
	// if " then read string
	if (line[linepos]=='"') {
		scString();
		return tok;		
	}


	if (!line[linepos]) return TT_EOL;
	// otherwise it must be a special character
	char ch = line[linepos++];
	lex = ch;
	switch (ch) {
		case '-':
			tok = TT_MINUS;
			break;

		case '+':
			tok = TT_PLUS;
			break;

		case '*':
			tok = TT_MULT;
			break;

		case '[':
			tok = TT_LB;
			break;

		case ']':
			tok = TT_RB;
			break;

		case '(':
			tok = TT_LP;
			break;

		case ')':
			tok = TT_RP;
			break;

		case '@':
			tok = TT_AT;
			break;

		case '.':
			tok = TT_DOT;
			break;

		case ',':
			tok = TT_COMMA;
			break;

		case '=':
			tok = TT_EQ;
			break;

		case ':':
			tok = TT_COLON;
			break;

		default:
			throw EScanner(EScanner::E_INVALIDCHAR, linepos);
	}
	return tok;
}


// Scan identifier, check for keyword
void Scanner::scId() {
	while ( line[linepos] && (isalnum(line[linepos]) || line[linepos]=='_') ) {
		lex = lex+line[linepos];
		linepos++;
	}
	DEBUG("'" << lex << "'	'" << line << "'");
	tok = TT_ID;
	if (chk_kw) check_keyword();
}


// Scan a string. The string, excluding "", is stored in lex
void Scanner::scString() {
	if (line[linepos]=='"') linepos++;
	while ( line[linepos] && (line[linepos]!='"') ) {
		if (line[linepos]=='\\') {
			char e;
			linepos++;
			switch (line[linepos]) {
				case 'b': e = 8; break;
				case 't': e = 9; break;
				case 'n': e = 10; break;
				case 'r': e = 13; break;
				case '"': e = '"'; break;
				case '\\': e = '\\'; break;
				default: throw EScanner(EScanner::E_UNKNOWNESC, linepos);
			}
			lex += e;
		} else
			lex += line[linepos];
		linepos++;
	}
	if (line[linepos]=='"') linepos++;
	else throw EScanner(EScanner::E_MISSINGQUOTE);
	DEBUG("'" << lex << "'	'" << line << "'");
	tok = TT_STRING;
}


// Scan value
void Scanner::scVal() {
	long base = 10;
	int digval = 0;
	val = 0;

	if (line[linepos]=='0') {
		linepos++;
		switch (line[linepos]) {
		case 'h': base=16; linepos++; break;
		case 'b': base=2; linepos++; break;
		case 'd': base=10; linepos++; break;
		case 'o': base=8; linepos++; break;
		default: base=10;
		}
	}

	while ( line[linepos] && isalnum(line[linepos]) ) {
		char curchar = line[linepos++];
		if (curchar>='0' && curchar<='9') digval=curchar-'0'; 
		if (curchar>='a' && curchar<='z') digval=curchar-'a'+10;
		if (digval>=base) throw EScanner(EScanner::E_DIGITLARGERBASE, linepos);
		val = val*base + digval;
	}
	tok = TT_CONST;
}


// check for keywords
void Scanner::check_keyword() {
	map<string,int>::iterator p = keywords.find(lex);
	if (p!=keywords.end())
		tok = static_cast<ScTokens>(p->second);

	if ( lex.length()==3 && lex[0]=='r' && isdigit(lex[1]) && isdigit(lex[2]) )
		tok = TT_REG;
}
