/////////////////////////////////////////////////////////////////////////////
//
// File: timer.h
// -------------
// Defines in-emulator timer system. The timer consist of a single class,
// Timer, containing functions for handling timers.
//
// Timers, in the context of R.1, is 32-bit counters working upwards toward
// the event value, at which an interrupt may be generated.
// +----------+
// | +------+ |
// | |   *=======* 
// | |   | adder |
// | |   *=======*  
// | |      | |
// | |   *=====================*
// | |   | comp. to event val. |
// | |   | if (eq and reset) 0 |---+
// | |   *=====================*   |
// | |      | |                  *===*
// | |   *=======*               | FF|
// | |   | REGIS |-- T?CLK       *===*
// | |   *=======*
// | +------+ |
// +----------+
//
#ifndef TIMER_H
#define TIMER_H

#include "R1emu.h"
#include "interrupt.h"

union Timer_control {
	struct {
		unsigned reserved: 26; 
		unsigned source: 2;    // reserved
		unsigned event: 1;     // bit set when timer ran out (eventval was reached)
		unsigned interrupt: 1; // generate interrupt when 0 reached
		unsigned reset: 1;     // reset when event is reached
		unsigned enabled: 1;   // the timer is enabled
	} flags;
	WORD word;
};

class Timer {
	public:
		// instantiate new timer linked to the interrupt-decoder IRQn
		Timer(IRQDecoder*, BYTE);
		
		// do a timer-tick. Returns value of done.
		void clk();
		
		// set event value. When counter reaches event value, done is true
		void setEvent( WORD );
		
		// set the control-word for this timer
		void setControl( WORD );
		
		// return current count value
		WORD getCount();
		
		// return control word
		WORD getControl();
		
		// event is set true when count reaches event value.
		// it is the users responsibility to clear the event by updating the 
		// control-word 
		bool getEvent();
		
	protected:
		WORD count;
		WORD eventval;
		Timer_control control;
		IRQDecoder* irq;
		BYTE irqno;
};

#endif